/**
 * Implemlementations for the grouse back end functions.  These
 * functions are all of the users actions with the QTwidgets.
 */
#include <qfiledialog.h>
#include <tulip/TlpTools.h>
#include <tulip/GlGraph.h>
#include <tulip/GlGraphWidget.h>
#include <tulip/ForEach.h>
#include "ColourShapeDefaults.h"
#include "GrouseInterfaceBackEnd.h"

using namespace std;
using namespace tlp;

const string finalLayoutLabel ("1__Final Layout");
const string finalLayoutDisplayLabel ("1__Final Layout Display");
const string hierarchyLabel ("1__Hierarchy");
const string hierarchyMetaGraphProperty ("viewMetaGraph");
const string viewLabel ("viewLabel");

//================================================================
/**
 * The function finds the first subgraph with this name in the
 * hierarchy.
 */
Graph *GrouseInterfaceBackEnd::findGraph (const string &graphName) {
  Graph *foundGraph = NULL, *curGraph;
  Iterator<Graph *> *subgraphs = rootGraph->getSubGraphs();
  while (subgraphs->hasNext()) {
    curGraph = subgraphs->next();
    string name = curGraph->getAttribute <string> ("name");
    if (name == graphName) {
      foundGraph = curGraph;
      break;
    }//end if
  } delete subgraphs; //end while
  return foundGraph;
}

//================================================================
void setGraphView (GlGraph *glGraph, Graph *newGraph) {
  GlGraphRenderingParameters param = glGraph->getRenderingParameters();
  param.setEdgeColorInterpolate (false);
  param.setEdgeSizeInterpolate (true);
  param.setBackgroundColor (BACKGROUND_COLOUR);
  param.setGraph (newGraph);
  glGraph->setRenderingParameters(param);
}//end setGraphView
//===============================================================
void GrouseInterfaceBackEnd::qtConnections () {
  connect (metaGraphHierarchyWidget1, 
	   SIGNAL (selectionChanged (const std::vector<tlp::node> &, 
				     const std::vector<tlp::Graph *> &)),
	   this,
	   SLOT (metaGraphHierarchyWidget1_selectionChanged (const std::vector<tlp::node> &, const std::vector<tlp::Graph *> &)));
  connect (metaGraphHierarchyWidget1,
	   SIGNAL (newPath (const std::vector<tlp::node> &)),
	   this,
	   SLOT (metaGraphHierarchyWidget1_newRightButtonPath (const std::vector<tlp::node> &)));
  connect (grouseSelectionProperty,
	   SIGNAL (selectionChanged (tlp::BooleanProperty &)),
	   	   this,
	   SLOT (graphWidgetSelectionChanged (tlp::BooleanProperty &)));
  connect (this,
	   SIGNAL (graphWidgetSelectionChangedSignal (tlp::BooleanProperty &)),
	   metaGraphHierarchyWidget1,
	   SLOT (selectionChanged (tlp::BooleanProperty &)));
 
}
//================================================================
void GrouseInterfaceBackEnd::initInterface (Graph *graph) {
  rootGraph = graph;
  //set up the metagraph hierarchy structure
  initHierarchy ();
  initFinalLayout();
  cerr << "loading meta-graph hierarchy" << endl;
  boundingCircles = new MetaNodeBoundingCircles (renderingWidget,
						 finalLayout);
  openCloseHierarchy = 
    new OpenCloseHierarchy (hierarchy, hierarchyRootNode,
			    finalLayout, boundingCircles, scaleMode);
  finalLayoutDisplay = openCloseHierarchy->getOpenCloseHierarchyGraph (); 
  metaGraphHierarchyWidget1
    ->setHierarchyProperties (hierarchyMetaGraphProperty,
			      viewLabel, hierarchy);
  grouseSelectionProperty = new GrouseSelectionProperty (finalLayoutDisplay,
							 hierarchy,
							 openCloseHierarchy);
  mouse = new MouseGrouseNavigate (openCloseHierarchy, 
				   metaGraphHierarchyWidget1,
				   grouseSelectionProperty);
  finalLayoutDisplay->setAttribute<string> ("name", finalLayoutDisplayLabel);
  renderingWidget->resetInteractors (mouse);
  setGraphView (renderingWidget, finalLayoutDisplay);
  qtConnections();
  renderingWidget->centerScene();
  renderingWidget->draw();
}

//================================================================
/**
 * This implements the action taken when the the open functions of
 * the file menu is selected.
 */
void GrouseInterfaceBackEnd::fileOpen() {
  QString s = QFileDialog::getOpenFileName("",
					   "Tlp (*.tlp *.tlp.gz)",
					   this,
					   "open file dialog",
					   "Choose a tulip file" );
  Graph *graph = tlp::loadGraph(s.latin1());
  initInterface (graph);
}
//================================================================
void GrouseInterfaceBackEnd::fileScale_Mode() {
  scaleMode = !scaleMode;
  openCloseHierarchy->setScaleMode (scaleMode);
}

//================================================================
void GrouseInterfaceBackEnd::fileLabelToggle() {
  GlGraphRenderingParameters param = renderingWidget->getRenderingParameters();
  labelToggle = !labelToggle;
  if (labelToggle) {
    param.setViewNodeLabel (true);
    param.setFontsType (1);
  }
  else {
    param.setViewNodeLabel (false);
  }
  renderingWidget->setRenderingParameters (param);
  renderingWidget->draw();
}

//================================================================

/**
 * Function which closes grouse
 */
void GrouseInterfaceBackEnd::fileExit() {
  exit(EXIT_SUCCESS);
}

//================================================================
void GrouseInterfaceBackEnd::fileSaveAs() {
  QString saveFile = QFileDialog::getSaveFileName("",
						  "Tlp (*.tlp *.tlp.gz)",
						  this,
						  "Save As dialog",
						  "Save As File name" );
   tlp::saveGraph (rootGraph, saveFile); 
}



//================================================================
/**
 * Sets up gouse bounding circles for the metanode hierarchy
 */
void GrouseInterfaceBackEnd::initHierarchy () {
  Graph *hierarchyRoot = findGraph (hierarchyLabel);
  if (hierarchyRoot != NULL) {
    GraphProperty *hierarchyMetaGraphs = 
      hierarchyRoot->getProperty<GraphProperty> (hierarchyMetaGraphProperty);
    hierarchyRootNode = hierarchyRoot->addNode();
    Iterator<Graph *> *firstLevel = hierarchyRoot->getSubGraphs();
    hierarchy = firstLevel->next();
    hierarchyMetaGraphs->setNodeValue (hierarchyRootNode, hierarchy);
    delete firstLevel;
  }//end if
}
//================================================================
/**
 * Set the rendering window to the final layout
 */
void GrouseInterfaceBackEnd::initFinalLayout () {
  finalLayout = findGraph (finalLayoutLabel);
}

//================================================================
void GrouseInterfaceBackEnd::
metaGraphHierarchyWidget1_newRightButtonPath (const vector<node> &newPath)
{
  StringProperty *labels = hierarchy->getProperty<StringProperty> (viewLabel);
  metaGraphHierarchyWidget1->clearSelection();
  for (int i = newPath.size() - 1; i >= 0; --i) {
    node hierNode = newPath[i];
    node groupNode = openCloseHierarchy->getDisplayNode (hierNode);
    //if the group node is visible and it is not a leaf
    if (groupNode.isValid() && (hierNode != groupNode)) {
      cout << "open node " << hierNode.id << endl;
      mouse->morphOpen (renderingWidget, groupNode);
      GlABoundingCircle *circle = 
	boundingCircles->getBoundingCircle (hierNode);
      const Color tlpColor = circle->getColor ();
      metaGraphHierarchyWidget1->setListViewColour (hierNode, tlpColor);
    }
    else
      cerr << "could not open " << labels->getNodeValue (hierNode) << endl;
  }//end for
}//end rightButton

//================================================================
void GrouseInterfaceBackEnd::
metaGraphHierarchyWidget1_selectionChanged(const vector<node> &hierNodes, 
					   const vector<Graph *> &hierGraphs){
  vector<GlABoundingCircle *> newSelBoundCircles;
  vector<node> newSelNodes;
  grouseSelectionProperty->clearSelection();
  for (unsigned int i = 0; i < hierNodes.size(); ++i) {
    GlABoundingCircle *circle = 
      boundingCircles->getBoundingCircle (hierNodes[i]);
    if (circle != NULL)
      newSelBoundCircles.push_back (circle);
    else if (hierNodes[i].isValid()) {
      //node has to be visible to be selected
      node displayNode = openCloseHierarchy->getDisplayNode (hierNodes[i]);
      if (displayNode.isValid() && finalLayoutDisplay->isElement (displayNode))
	newSelNodes.push_back (displayNode);
    }//end else if
  }//end for
  grouseSelectionProperty->setSelection (newSelNodes, newSelBoundCircles,
					 true);
  renderingWidget->draw();
}

//================================================================
void GrouseInterfaceBackEnd::
graphWidgetSelectionChanged (tlp::BooleanProperty &selection) {
  emit graphWidgetSelectionChangedSignal (selection);
}//end graphWidgetSelectionChanged
