#include <tulip/GraphProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/ForEach.h>
#include <tulip/GlGraph.h>
#include <tulip/Graph.h>
#include "ColouredListViewItem.h"
#include "MetaGraphHierarchyWidget.h"

using namespace std;
using namespace tlp;

const string viewLabel ("viewLabel");
const string viewMetaGraph ("viewMetaGraph");
const QColor defaultColour (255, 255, 255);
QListViewItem *selItem = NULL;

//=======================================================
class MetaGraphListViewItem : public ColouredListViewItem {
private:
  tlp::node hierNode;
public:
  MetaGraphListViewItem (QListView *parent) : 
    ColouredListViewItem(parent)
    {}

  MetaGraphListViewItem (QListViewItem *item,
			 const tlp::node &hierNode) : 
    ColouredListViewItem(item), hierNode(hierNode) {}

  /**
   * A function to return the hierarchy node associated with this
   * coloured list view item.
   */
  tlp::node getHierarchyNode () { return hierNode; }
};

//=======================================================================
MetaGraphHierarchyWidget::
MetaGraphHierarchyWidget (QWidget *parent,
			  const char *name, WFlags flags, 
			  tlp::Graph *hierarchy,
			  const string &metagraphProperty,
			  const string &labelProperty) :
  MetaGraphHierarchyWidgetData (parent, name, flags),
  hierarchy(hierarchy), 
  metagraphProperty(metagraphProperty),
  labelProperty(labelProperty)
{
  hierNodeToListViewItem.setAll (NULL);
  populate();
}

//=======================================================================
MetaGraphHierarchyWidget::
MetaGraphHierarchyWidget (QWidget* parent, const char *name, Qt::WFlags fl) :
  MetaGraphHierarchyWidgetData (parent, name, fl),
  metagraphProperty(viewMetaGraph),
  labelProperty(viewLabel)
{
  hierNodeToListViewItem.setAll (NULL);
}
  

//=======================================================================
MetaGraphHierarchyWidget::~MetaGraphHierarchyWidget () {
  metaGraphHierarchyListView->clear();
}

//=======================================================================
node MetaGraphHierarchyWidget::getHierarchyNode (QListViewItem *listViewItem) {
  return ((MetaGraphListViewItem *)listViewItem)->getHierarchyNode();
}//end getHierarchyNode

//=======================================================================
Graph *MetaGraphHierarchyWidget::getHierarchyGraph (QListViewItem 
						    *listViewItem) {
  GraphProperty *metaGraphHierarchy = 
    hierarchy->getProperty<GraphProperty> (metagraphProperty);
  node queryNode = getHierarchyNode (listViewItem);
  if (queryNode.isValid())
    return metaGraphHierarchy->getNodeValue (queryNode);
  else if (listViewItem->parent() == NULL)
    return hierarchy;
  else
    return NULL;
}

//=======================================================================
void MetaGraphHierarchyWidget::
setHierarchyProperties (const string &newMetaProperty,
			const string &newLabelProperty,
			Graph *newRootGraph) {
  metagraphProperty = newMetaProperty;
  labelProperty = newLabelProperty;
  hierarchy = newRootGraph;
  populate();
}

//=======================================================================
void 
MetaGraphHierarchyWidget::setMetaGraphProperty (const string &newProperty) {
  metagraphProperty = newProperty;
  metaGraphHierarchyListView->clear();
  populate();
}

//=======================================================================
void 
MetaGraphHierarchyWidget::setLabelProperty (const string &newProperty) {
  labelProperty = newProperty;
  metaGraphHierarchyListView->clear();
  populate();
}

//=======================================================================
void 
MetaGraphHierarchyWidget::setHierarchyGraph (Graph *rootGraph) {
  hierarchy = rootGraph;
  metaGraphHierarchyListView->clear();
  populate();
}

//=======================================================================
void MetaGraphHierarchyWidget::setSelectedLabel (const node &hierarchyNode) {
  QListViewItem *selectedItem = 
    hierNodeToListViewItem.get (hierarchyNode.id);
  if (selectedItem != NULL) {
    ((MetaGraphListViewItem *)selectedItem)->setSelected (true);
    metaGraphHierarchyListView->repaintItem (selectedItem);
  }//end if
}

//=======================================================================
 void 
 MetaGraphHierarchyWidget::setListViewColour (const node &hierarchyNode,
					      const Color &nodeColour) {
    ColouredListViewItem *clistViewItem =
      ((ColouredListViewItem *)hierNodeToListViewItem.get (hierarchyNode.id));
    QColor qboundColor (nodeColour.getR(), nodeColour.getG(), 
			nodeColour.getB());
    clistViewItem->setBackground(0, qboundColor);
    metaGraphHierarchyListView->repaintItem (clistViewItem);
 }

//=======================================================================
void MetaGraphHierarchyWidget::populate () {
    metaGraphHierarchyListView->clear();
    metaGraphHierarchyListView->setColumnText (0, "MetaGraph Hierarchy");
    MetaGraphListViewItem *root = 
      new MetaGraphListViewItem (metaGraphHierarchyListView);
    root->setOpen (true);
    std::string name = hierarchy->getAttribute<std::string> ("name");
    root->setText (0, QString (name));
    root->setBackground(0, defaultColour);
    populate (hierarchy, root);
  }
//=======================================================================
void MetaGraphHierarchyWidget::populate (Graph *curGraph,
					 QListViewItem *inputItem) {
  MetaGraphListViewItem *curItem =
    static_cast<MetaGraphListViewItem *> (inputItem);
  GraphProperty *viewMetaGraph = 
    curGraph->getProperty<GraphProperty> (metagraphProperty);
  StringProperty *viewLabel = 
    curGraph->getProperty<StringProperty> (labelProperty);
  node curNode;
  forEach (curNode, curGraph->getNodes()) {
    Graph *metaGraph = viewMetaGraph->getNodeValue (curNode);
    MetaGraphListViewItem *newItem = 
      new MetaGraphListViewItem (curItem, curNode);
    hierNodeToListViewItem.set (curNode.id, newItem);
    //if this is a meta node, title it the same as its graph
    if (metaGraph != NULL) {
      newItem->setOpen (true);
      std::string name = metaGraph->getAttribute<std::string> ("name");
      newItem->setText (0, QString(name));
      newItem->setBackground(0, defaultColour);
      populate (metaGraph, newItem);
    }
    //otherwise a leaf of the hierarchy (original node of graph)
    else {
      newItem->setOpen (true);
      string label = viewLabel->getNodeValue (curNode);
      newItem->setText (0, QString(label));
      newItem->setBackground(0, defaultColour);
    }
  }
}

//=======================================================================
void MetaGraphHierarchyWidget::clearSelection () {
  metaGraphHierarchyListView->clearSelection();
  metaGraphHierarchyListView->update();
}

//=======================================================================
void MetaGraphHierarchyWidget::selectAll () {
  metaGraphHierarchyListView->clearSelection();
  metaGraphHierarchyListView->invertSelection();
  metaGraphHierarchyListView->update();
}

//=======================================================================
void MetaGraphHierarchyWidget::
selectionChanged (BooleanProperty &newSelection) {
  metaGraphHierarchyListView->setSelectionMode (QListView::Single);
  if (selItem != NULL) {
    selItem->setSelected (false);
    selItem->repaint();
  }
  metaGraphHierarchyListView->update();
  Iterator<node> *newNodeSelection = newSelection.getNodesEqualTo (true);
  while (newNodeSelection->hasNext()) {
    selItem = hierNodeToListViewItem.get (newNodeSelection->next().id);
    selItem->setSelected (true);
  } delete newNodeSelection; //end while
  metaGraphHierarchyListView->update();
}//end selectionChanged

//=======================================================================
void MetaGraphHierarchyWidget::selectionUpdate (vector<node> &nodeSelection,
						vector<Graph *> &graphSelection) {
  metaGraphHierarchyListView->update();
  QListViewItemIterator listViewIterator (metaGraphHierarchyListView);
  while (listViewIterator.current() != NULL) {
    QListViewItem *curItem = listViewIterator.current();
    if (curItem->isSelected()) {
      nodeSelection.push_back (getHierarchyNode (curItem));
      graphSelection.push_back (getHierarchyGraph (curItem));
    }//end if selected
    ++listViewIterator;
  }//end while
}//end selectionUpdate

//=======================================================================
void MetaGraphHierarchyWidget::newHierarchyPath (QListViewItem *pathOrigin,
						 vector<node> &path) {
  QListViewItem *curItem = pathOrigin;
  while (curItem->parent() != NULL) {
    path.push_back (getHierarchyNode (curItem));
    curItem = curItem->parent();
  }//end while
  metaGraphHierarchyListView->clearSelection();
  metaGraphHierarchyListView->update();
}

//=======================================================================
void MetaGraphHierarchyWidget::
metaGraphHierarchyListView_mouseButtonClicked(int button, 
					      QListViewItem *itemClicked, 
					      const QPoint &pointClicked, 
					      int column) {
  if (itemClicked == NULL) return;
  if (button ==Qt::LeftButton) {
    vector<node> nodeSelection;
    vector<Graph *> graphSelection;
    nodeSelection.push_back (getHierarchyNode (itemClicked));
    graphSelection.push_back (getHierarchyGraph (itemClicked));
    //selectionUpdate (nodeSelection, graphSelection);
    emit selectionChanged (nodeSelection, graphSelection);
  }
  else if (button == Qt::RightButton) {
    vector<node> path;
    newHierarchyPath (itemClicked, path);
    emit newPath (path);
  }
}

