
#include <string>
#include <vector>
#include <qlistview.h>
#include <tulip/Graph.h>
#include <tulip/StringProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/MutableContainer.h>
#include "MetaGraphHierarchyWidgetData.h"

#ifndef __META_GRAPH_HIERARCHY_CLASS__
#define __META_GRAPH_HIERARCHY_CLASS__

class MetaGraphHierarchyWidget : public MetaGraphHierarchyWidgetData {
  Q_OBJECT;
 public:
  MetaGraphHierarchyWidget (QWidget *parent,
			    const char *name, WFlags flags, 
			    tlp::Graph *hierarchy,
			    const std::string &metagraphProperty,
			    const std::string &labelsProperty);
  MetaGraphHierarchyWidget (QWidget* parent = 0, const char *name = 0, 
			    Qt::WFlags fl = 0);

  ~MetaGraphHierarchyWidget ();

  /**
   * Returns the node associated with this listViewItem
   * \param listViewItem MetaGraphHierarchyListView item which we
   * want to retrieve the corresponding node.
   */
  tlp::node getHierarchyNode (QListViewItem *listViewItem);

  /**
   * Returns graph in the hierarchy associated with this listViewItem
   * \param listViewItem MetaGraphHierarchyListView item which we
   * want to retrieve the corresponding graph.
   */
  tlp::Graph *getHierarchyGraph (QListViewItem *listViewItem);

  /**
   * Sets the metagraph, label, and root of the hierarchy for this
   * graph widget.
   * \param newMetaProperty The name of the metagraph property.
   * \param newLabelProperty The name of the label property.  This determines
   * the names in the metagraph hierarchy list view widget.
   * \param newRootGraph The root graph of the widget.
   */
  void setHierarchyProperties (const std::string &newMetaProperty,
			       const std::string &newLabelProperty,
			       tlp::Graph *newRootGraph);

  /**
   * Sets the metaGraph property for which this widget represents
   * \param newProperty The new metaGraph property this widget represents.
   */
  void setMetaGraphProperty (const std::string &newProperty);

  /**
   * Sets the label property.  This property is used to represent leaves
   * of the metagraph hierarchy in the widget.
   * \param newProperty The new metaGraph property this widget represents.
   */
  void setLabelProperty (const std::string &newProperty);

  /**
   * Sets the root graph which the widget will use to populate its entries.
   * \param rootGraph The new root graph of the metagraph hierarchy
   * this widget represents.
   */
  void setHierarchyGraph (tlp::Graph *rootGraph);

  /**
   * Given a node in the graph hierarchy.  This node sets the
   * label in the MetaGraphHierarchy to selected.
   * \param hierarchyNode The node we want to select.
   */
  void setSelectedLabel (const tlp::node &hierarchyNode);

  /**
   * Given a node in the graph hierarchy.  This node sets the
   * label in the MetaGraphHierarchy to the specified colour.
   * \param hierarchyNode The node for which we want to change label colour
   * \param nodeColour The new colour of the label.
   */
  void setListViewColour (const tlp::node &hierarchyNode,
			  const tlp::Color &nodeColour);

  /**
   * Function to populate the MetaGraphHierarchyWidget.  It will
   * use the hierarchy graph and proxies alread associated with the
   * widget to do its work.
   */
  void populate ();

  /**
   * Unselects all elements of the listview.
   */
  void clearSelection ();

  /**
   * Selects all elements of the listview.
   */
  void selectAll ();
 
  
 signals:
  /**
   * Signal emitted when selection in the widget has changed.
   * \param newNode The new node in the meta graph hierarchy.
   * \param newHierGraph The new hierarchy subgraph in the hierarchy.
   */
  void selectionChanged (const std::vector<tlp::node> &newNodes, 
			 const std::vector<tlp::Graph *> &newHierGraphs);

  /**
   * Signal emitted when a new right button path through the hierarchy
   * has been specified.
   * \param newPath The new path through the hierarchy.
   */
  void newPath (const std::vector<tlp::node> &nodePath);

  public slots:
  /**
   * Slot to update the list view with the new passed selection.
   */
    void selectionChanged (tlp::BooleanProperty &newSelection);

 private:
  tlp::Graph *hierarchy;
  std::string metagraphProperty;
  std::string labelProperty;
  tlp::MutableContainer<QListViewItem *> hierNodeToListViewItem;

  /**
   * Function to populate the MetaGraphHierarchyWidget. The private
   * method considers a graph in the hierarchy and its associated item.
   */
  void populate (tlp::Graph *curGraph, QListViewItem *inputItem);

  /**
   * Function to handle when the selection has been updated.
   * \param nodeSelection the set of hierarchy nodes in the selection
   * \param graphSelection the set of graphs in the selection
   */
  void selectionUpdate (std::vector<tlp::node> &nodeSelection,
			std::vector<tlp::Graph *> &graphSelection);

  /**
   * Function to handle when a new hierarchy path is requested
   * \param pathOrigin The item where we start to trace back up the path
   * \param path The returned path by the function.
   */
  void newHierarchyPath (QListViewItem *pathOrigin, 
			 std::vector<tlp::node> &path);

  /**
   * Signal handler for whenever a mouse button is clicked on an element of
   * the listview.
   * \param the QT button which was clicked.
   * \param newItem The new item which is selected.
   * \param coord The new point.
   * \param col the column of the listview
   */
  void metaGraphHierarchyListView_mouseButtonClicked(int button,
						     QListViewItem *newItem, 
						     const QPoint &coord, 
						     int col);

 
};
#endif
