/**
 * This file implements metanode highlighting using GlGraphWidgets.
 * For each metanode, it looks at the boundary circle and renders it as
 * a gl widget.
 */
#include <cmath>
#include <iostream>
#include <tulip/ForEach.h>
#include <tulip/GraphProperty.h>
#include <tulip/DoubleProperty.h>
#include "HierarchyMap.h"
#include "MetaNodeBoundingCircle.h"
using namespace std;
using namespace tlp;

class HSV {
public:
  float h;
  float s;
  float v;
  HSV (float h, float s, float v) {
    this->h = h;
    this->s = s;
    this->v = v;
  }
};

//==============================================
#define UNDEFINED (0) 
HSV rgb_to_hsv (const Color &rgb) {
  float R = rgb.getR()/255.0;
  float G = rgb.getG()/255.0;
  float B = rgb.getB()/255.0;
  float v, x, f;
  int i;
  x = R; v = R;
  if (x > G) x = G;
  if (x > B) x = B;
  if (v < G) v = G;
  if (v < B) v = B;
  if(v == x) return HSV(UNDEFINED, 0, v);
  f = (fabs(R - x) < 1e-6) ? G - B : (fabs (G - x) < 1e-6 ? B - R : R - G);
  i = (fabs(R - x) < 1e-6) ? 3 : (fabs(G - x) < 1e-6 ? 5 : 1);
  return HSV(i - f /(v - x), (v - x)/v, v);
}
//==============================================
inline unsigned char toChar (float value) {
  return static_cast<unsigned char> (floor (value * 255));
}
Color hsv_to_rgb (const HSV &hsv) {

  // H is given on [0, 6] or UNDEFINED. S and V are given on [0, 1].
  // RGB are each returned on [0, 1].
  float h = hsv.h, s = hsv.s, v = hsv.v, m, n, f;
  int i;
  if(h == UNDEFINED) return Color (toChar(v), toChar(v), toChar(v));

  i = floor(h);
  f = h - i;
  if(!(i & 1)) f = 1 - f; // if i is even
  m = v * (1 - s);
  n = v * (1 - s * f);
  switch (i) {
  case 6:
  case 0: return Color(toChar(v), toChar(n), toChar(m));
  case 1: return Color(toChar(n), toChar(v), toChar(m));
  case 2: return Color(toChar(m), toChar(v), toChar(n));
  case 3: return Color(toChar(m), toChar(n), toChar(v));
  case 4: return Color(toChar(n), toChar(m), toChar(v));
  case 5: return Color(toChar(v), toChar(m), toChar(n));
  }
}

//base depth for bouding objects with initialization
double GlABoundingCircle::baseDepth = -1.0;

#define LOWER_CIRCLE_BY (5)
//================================================================
unsigned int 
computeDepthByLevel (Graph *displayGraph, Graph *displaySubGraph) {
  double depthLevel = 0;
  Graph *curGraph = displaySubGraph;
  while (curGraph != displayGraph) {
    curGraph = curGraph->getSuperGraph();
    ++depthLevel;
  }//end while
  return depthLevel;
}

//================================================================
unsigned int 
MetaNodeBoundingCircles::computeMaxDepth (Graph *finalLayout) {
  unsigned int curMaxDepth = 0;
  Iterator<Graph *> *subgraphs = finalLayout->getSubGraphs();
  while (subgraphs->hasNext()) {
    Graph *curGraph = subgraphs->next();
    unsigned int subGraphDepth = 1 + computeMaxDepth (curGraph);
    if (curMaxDepth < subGraphDepth) 
      curMaxDepth = subGraphDepth;
  } delete subgraphs;
  return curMaxDepth;
}//end computeMaxDepth

//================================================================
/**
 * Given a bounding circle size object, it draws the rendering widget
 * in the tulip window.
 */
void 
MetaNodeBoundingCircles::addBoundingCircle (Graph *displayGraph,
					    Graph *displaySubGraph,
					    const node &hierarchyNode,
					    const Coord &circleCentre, 
					    const Size &circleSize, 
					    const Color& circleColour) {

  if (hierNodeToCircle.find (hierarchyNode) != hierNodeToCircle.end()) {
    cerr << "Subgraph already has a bounding circle " << endl;
    return;
  }
  char charID[10];
  sprintf (charID, "%d", hierarchyNode.id);
  string boundCircleID (charID);

  //recheck the baseDepth
  LayoutProperty *viewLayout = 
    displayGraph->getProperty<LayoutProperty> ("viewLayout");
  edge curEdge;
  forEach (curEdge, displayGraph->getEdges()) {
    const vector<Coord> &bends = viewLayout->getEdgeValue (curEdge);
    for (unsigned int i = 0; i < bends.size(); ++i) {
      if (bends[i].getZ() < GlABoundingCircle::getBaseDepth ())
	GlABoundingCircle::setBaseDepth (bends[i].getZ());
    }//end for
  }//end forEach

  //compute the depth by level
  unsigned int levelDepth = 
    computeDepthByLevel (displayGraph, displaySubGraph);

  HSV circleHSV = rgb_to_hsv (circleColour);
  circleHSV.s = circleHSV.s*(0.3 + 0.7*(levelDepth/((double)maxDepth)));
  circleHSV.v += 0.1;
  Color washedColor = hsv_to_rgb (circleHSV);
  
  double depth = static_cast<double> (levelDepth) - 
    static_cast<double> (maxDepth);
  hierNodeToCircle[hierarchyNode] = 
    new GlABoundingCircle (hierarchyNode, displaySubGraph,
			   circleCentre, circleSize, 
			   washedColor, depth);
  showBoundingCircle (hierarchyNode);
}
//================================================================
void GlABoundingCircle::computePoints () {
  points.clear();
  double alpha = 0;
  const double delta = 2.0*M_PI/30.0;
  const double rad = sqrt (size.getW()*size.getW()/4.0 + 
			   size.getH()*size.getH()/4.0);
  Coord curPoint;
  for(int i = 0; i <= 30; ++i) {
    curPoint.setX (rad*cos(alpha) + pos.getX());
    curPoint.setY (rad*sin(alpha) + pos.getY());
    curPoint.setZ (depth + baseDepth);
    points.push_back (curPoint);
    alpha += delta;
  }//end for
}//end computePoints

//================================================================
 void GlABoundingCircle::draw (GlGraph *graphRenderer) {
   setup();
   glColor4fv(color.getGL());
   glBegin(GL_TRIANGLE_FAN);
   glVertex3d (pos.getX(), pos.getY(), depth);
   for(unsigned int i=0; i < points.size(); ++i)
     glVertex3d(points[i].getX(), points[i].getY(), points[i].getZ());
   glEnd();
    if (!selected) {
      glColor4fv (META_NODE_BORDER_COLOUR.getGL());
      glLineWidth (3.0);
    }
    else {
      glColor4fv (SELECTION_COLOUR.getGL());
      glLineWidth (8.0);
    }
    glBegin(GL_LINES);
    for(unsigned int i=0; i < points.size() - 1; ++i) {
      glVertex3d(points[i].getX(), points[i].getY(), points[i].getZ());
      glVertex3d(points[i + 1].getX(), points[i + 1].getY(), 
		 points[i + 1].getZ());
    }//end for
    glEnd();
    unsetup();
 }//end draw
