/**
 * This file implements metanode highlighting using GlGraphWidgets.
 * For each metanode, it looks at the boundary circle and renders it as
 * a gl widget.
 */

#ifndef __META_NODE_BOUNDING_CIRCLE__
#define __META_NODE_BOUNDING_CIRCLE__
#include <cstdio>
#include <vector>
#include <string>
#include <tulip/GlAugmentedDisplay.h>
#include <tulip/GlGraphWidget.h>
#include <tulip/GlGraph.h>
#include <tulip/ForEach.h>
#include <tulip/Graph.h>
#include <tulip/SizeProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ColorProperty.h>
#include "ColourShapeDefaults.h"

class GlABoundingCircle : public tlp::GlAugmentedDisplay {
 private:
  tlp::node hierarchyNode;
  tlp::Graph *displaySubGraph;
  tlp::Size size;
  tlp::Coord pos;
  tlp::Color color;
  double depth;
  static double baseDepth;
  bool selected;
  std::vector<Coord> points;

 public:

  void computePoints (); 

  GlABoundingCircle (tlp::node hierarchyNode, tlp::Graph *displaySubGraph,
		     Coord pos, Size size, Color color, double depth) :
    hierarchyNode(hierarchyNode), displaySubGraph(displaySubGraph),
    pos(pos), size(size), color(color), 
    depth(depth), selected (false) {
    computePoints();
  }

  inline void update (const tlp::Coord &newPos,
		      const tlp::Size &newSize,
		      double newDepth) {
    pos = newPos;
    size = newSize;
    depth = newDepth;
    computePoints();
  }

  const std::vector<Coord> &getPoints () {
    return points;
  }

  void setPoints (const tlp::Coord &newPos,
		  const std::vector<Coord> &newPoints) {
    pos = newPos;
    points = newPoints;
  }

 void draw (GlGraph *graphRenderer);
 
  inline bool getSelected () { return selected; }
  inline double getDepth () { return depth; }
  inline static double getBaseDepth () { return baseDepth; } 
  inline const Color &getColor () { return color; }
  inline const Coord &getPos () { return pos; }
  inline const Size &getSize () { return size; }
  inline double getRadius () { return sqrt (size.getW()*size.getW()/4.0 + 
					    size.getH()*size.getH()/4.0); }
  inline const tlp::node &getHierarchyNode () { return hierarchyNode; }
  inline tlp::Graph *getDisplaySubGraph () { return displaySubGraph; }
  inline static void setBaseDepth (double newBaseDepth) { 
    baseDepth = newBaseDepth;
  } 
  inline void setSelected (bool newSel) { selected = newSel; }
};

class MetaNodeBoundingCircles {
 private:
  GlGraphWidget *renderingWidget;
  stdext::hash_map<tlp::node, GlABoundingCircle*> hierNodeToCircle;
  unsigned int maxDepth;

  /**
   * Function to compute the maximum depth of the hierarchy
   */
  unsigned int computeMaxDepth (Graph *graph);

  /**
   * Function to return the string id of a bounding object given its
   * hierarchy node
   */
  inline std::string getID (const tlp::node &hierNode) {
    char charID[10];
    std::sprintf (charID, "%d", hierNode.id);
    return charID;
  }
  
 public:
  MetaNodeBoundingCircles (GlGraphWidget *renderingWidget,
			   Graph *finalLayout) :
    renderingWidget(renderingWidget) {
    maxDepth = computeMaxDepth (finalLayout);
  }

  inline ~MetaNodeBoundingCircles () {
    stdext::hash_map<tlp::node, GlABoundingCircle*>::iterator nodeCircleIt =
      hierNodeToCircle.begin();
    for (; nodeCircleIt != hierNodeToCircle.end(); ++nodeCircleIt)
      delete nodeCircleIt->second;
  }//end destructor

  /**
   * Given a bounding circle size object, it draws the rendering widget
   * in the tulip window.
   */
  void addBoundingCircle (tlp::Graph *displayGraph,
			  tlp::Graph *displaySubGraph,
			  const tlp::node &hierarchyNode,
			  const tlp::Coord &circleCentre, 
			  const tlp::Size &circleSize, 
			  const tlp::Color &circleColour);
  
  /**
   * Given a hierarchy node, this function returns the bounding circle
   */
  inline GlABoundingCircle *getBoundingCircle (const tlp::node &hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end())
      return NULL;
    else
      return hierNodeToCircle[hierNode];
  }

  /**
   * Given a hierarchy node, this function hides the bounding circle
   */
  inline void hideBoundingCircle (const tlp::node hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end()) {
      cerr << "cannot hide circle of hierarchy node " << hierNode.id << endl 
	   << "It is not present." << endl;
      return;
    }
    const std::string stringID = getID (hierNode);
    renderingWidget->removeGlAugmentedDisplay (stringID.c_str());
  }//end remove

  /**
   * Given a hierarchy node, this function shows the bounding circle
   */
  inline void showBoundingCircle (const tlp::node &hierNode) {
    if (hierNodeToCircle.find(hierNode) == hierNodeToCircle.end()) {
      cerr << "cannot show circle of hierarchy node " << hierNode.id << endl 
	   << "It is not created." << endl;
      return;
    } 
    
    const std::string stringID = getID (hierNode);
    renderingWidget
      ->addGlAugmentedDisplay (hierNodeToCircle.find(hierNode)->second, 
			       stringID.c_str());
  }
};
#endif
