#include <queue>
#include <tulip/BooleanProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/SizeProperty.h>
#include "MouseGrouseNavigate.h"

using namespace tlp;
using namespace std;

//===============================================================
//Defines for buttons to be used on each mouse operation
#define QT_FREE_PAN_BUTTON (Qt::LeftButton)
#define QT_FREE_PAN_KEY_DOWN (Qt::ControlButton)
#define QT_SELECT_BUTTON (Qt::LeftButton)
#define QT_SELECT_KEY_DOWN (Qt::NoButton)
#define QT_FREE_PAN_DRAG_STATE (QT_FREE_PAN_BUTTON | QT_FREE_PAN_KEY_DOWN)
#define QT_OPEN_CLOSE_BUTTON (Qt::RightButton)
#define QT_OPEN_CLOSE_KEY_DOWN (Qt::NoButton)
#define QT_POWER_OPEN_BUTTON (Qt::RightButton)
#define QT_POWER_OPEN_KEY_DOWN (Qt::ShiftButton)

//===============================================================

//===============================================================
/*
 * Default constructors and destructors
 */
MouseGrouseNavigate::MouseGrouseNavigate (OpenCloseHierarchy 
					  *openCloseHierarchy,
					  MetaGraphHierarchyWidget 
					  *metaGraphHierarchy,
					  GrouseSelectionProperty
					  *grouseSelectionProperty) :
  openCloseHierarchy (openCloseHierarchy), 
  metaGraphHierarchy(metaGraphHierarchy),
  grouseSelectionProperty(grouseSelectionProperty) {}
MouseGrouseNavigate::~MouseGrouseNavigate() {}

//===============================================================
//This local function picks the top augmented display
GlAugmentedDisplay *pickTopCircle (vector <pair <string, GlAugmentedDisplay *> > &pickedAug) {
  vector <pair <string, GlAugmentedDisplay *> >::iterator pickedAugIt =
    pickedAug.begin();
  GlABoundingCircle *curCircle = 
    static_cast<GlABoundingCircle *> (pickedAugIt->second);
  GlABoundingCircle *topCircle = curCircle;
  double topDepth = curCircle->getDepth();
  for (++pickedAugIt; pickedAugIt != pickedAug.end(); ++pickedAugIt) {
    curCircle = static_cast<GlABoundingCircle *> (pickedAugIt->second);
    if (topDepth < curCircle->getDepth()) {
      topCircle = curCircle;
      topDepth = curCircle->getDepth();
    }//end if
  }//end for
  return topCircle;
}//end pickTopCircle

//===============================================================
/**
 * Private function for selecting graph elements.  x and y are the
 * click points, glW is the graph widget, and if reset is set to 
 * true, all elements are deselected before selection.
 */
void pick (unsigned int x, unsigned int y, GlGraphWidget *glW, bool reset,
	   bool resetAfter,
	   bool pickNodes, bool pickEdges, bool pickAugs,
	   node &topNode, edge &topEdge, GlAugmentedDisplay *&topAug) {
  if (!(pickNodes || pickEdges || pickAugs)) return;
  Observable::holdObservers();
  bool result;
  ElementType type;

  node tmpNode;
  edge tmpEdge;
  BooleanProperty *elementSelected = 
    glW->getGraph()->getProperty<BooleanProperty>("viewSelection");
  if (reset) {
    elementSelected->setAllNodeValue(false);
    elementSelected->setAllEdgeValue(false);
  }
  
  if (pickNodes || pickEdges) {
    result = glW->doSelect(x, y, type, tmpNode, tmpEdge);
    if (result) {
      switch(type) {
      case NODE:
	if (pickNodes) {
	  elementSelected
	    ->setNodeValue(tmpNode, !elementSelected->getNodeValue(tmpNode));
	  topNode = tmpNode;
	}
	break;
      case EDGE:
	if (pickEdges) {
	  elementSelected
	    ->setEdgeValue(tmpEdge, !elementSelected->getEdgeValue(tmpEdge));
	  topEdge = tmpEdge;
	}
	break;
      }
    }
  }//end if

  if (pickAugs) {
    vector <pair <string, GlAugmentedDisplay *> > pickedAugmentedDisplays;
    result = glW->doSelectAugmentedDisplay (x, y, pickedAugmentedDisplays);
    if (result) topAug = pickTopCircle (pickedAugmentedDisplays);
    else topAug = NULL;
  }
  if (resetAfter) {
    elementSelected->setAllNodeValue(false);
    elementSelected->setAllEdgeValue(false);
  }
  Observable::unholdObservers();
}

//===============================================================
/**
 * Function to pan the mouse.
 */
void MouseGrouseNavigate::pan (const QPoint &end, GlGraphWidget *graph) {
  if (end.x() != startDrag.x()) 
    graph->translateCamera(end.x() - startDrag.x(), 0, 0);
  if (end.y() != startDrag.y()) 
    graph->translateCamera(0, startDrag.y() - end.y(), 0);
  initStartDrag (end);
}

//===============================================================
/**
 * Function to select a internal node of the hierarchy
 */
void MouseGrouseNavigate::select (const QPoint &mouseClick, 
				  GlGraphWidget *graphWidget) {
  node pickedNode;
  edge pickedEdge;
  vector<node> nodeSelections;
  vector<GlABoundingCircle *> circleSelections;
  GlAugmentedDisplay *pickedAug = NULL;
  grouseSelectionProperty->clearSelection();
  metaGraphHierarchy->clearSelection();
  pick (mouseClick.x(), mouseClick.y(), graphWidget, true, true, true, 
	false, true, pickedNode, pickedEdge, pickedAug);

  node hierarchyNode;
  if (pickedNode.isValid())
    hierarchyNode = openCloseHierarchy->getHierarchyNode (pickedNode);
  if (hierarchyNode.isValid()) {
    StringProperty *viewLabels =
      graphWidget->getGraph()->getProperty<StringProperty> ("viewLabel");
    metaGraphHierarchy->setSelectedLabel (hierarchyNode);
    nodeSelections.push_back (pickedNode);
  }
  else if (pickedAug != NULL) {
    GlABoundingCircle *boundingCircle =
      static_cast<GlABoundingCircle *> (pickedAug);
    metaGraphHierarchy->setSelectedLabel (boundingCircle->getHierarchyNode());
    circleSelections.push_back (boundingCircle);
  }
  
  if (!nodeSelections.empty() || !circleSelections.empty()) {
    grouseSelectionProperty->setSelection (nodeSelections, circleSelections);
    graphWidget->draw();
  }//end selection changed
}

//===============================================================
void fixForAdditions (const vector<node> &nodesAdded,
		      const vector<edge> &edgesAdded,
		      const vector<GlABoundingCircle *> &augsAdded,
		      stdext::hash_map<node, unsigned int> &order,
		      const Coord &startPos,
		      GraphState *gi, GraphState *gf,
		      GlGraphWidget *glW) {
  ColorProperty *colours = 
    glW->getGraph()->getProperty<ColorProperty> ("viewColor");
  SizeProperty *sizes = 
    glW->getGraph()->getProperty<SizeProperty> ("viewSize");
  for (unsigned int i = 0; i < nodesAdded.size(); ++i) {
    gi->layout->setNodeValue (nodesAdded[i], startPos);
    gi->size->setNodeValue (nodesAdded[i],
			    sizes->getNodeValue (nodesAdded[i]));
    gi->color->setNodeValue (nodesAdded[i],
			     colours->getNodeValue (nodesAdded[i]));
  }//end for nodesAdded
  for (unsigned int i = 0; i < edgesAdded.size(); ++i) {
    gi->layout->setEdgeValue (edgesAdded[i], vector<Coord> ());
    gi->color->setEdgeValue (edgesAdded[i],
			     colours->getEdgeValue (edgesAdded[i]));
  }//end for edgesAdded

  //Adjust for aug display insertions or deletions
  if (gi->augPoints.size() < gf->augPoints.size()) {
    gi->augPoints.resize (gf->augPoints.size());
    gi->augCenters.resize (gf->augPoints.size());
  }//end if
  for (unsigned int i = 0; i < augsAdded.size(); ++i) {
    vector<Coord> addedDisplay;
    node hierarchyNode = augsAdded[i]->getHierarchyNode();
    const vector<Coord> &points = augsAdded[i]->getPoints(); 
    for (unsigned int j = 0; j < points.size(); ++j)
      addedDisplay.push_back (startPos); 

    unsigned int place = order[hierarchyNode];
    gi->augPoints[place] = addedDisplay;
    gi->augCenters[place] = startPos;
  }
}

//===============================================================
void updateAugs (GlGraphWidget *graphWidget, GraphState *g0, 
		 stdext::hash_map<node, unsigned int> &order) {
  const stdext::hash_map<std::string, GlAugmentedDisplay*> *augDisplays =
    graphWidget->getAugmentedDisplays();
  if ((augDisplays != NULL) && !g0->interpolateCenters.empty()
       && !g0->curInterpolation.empty()) {
    stdext::hash_map<string, GlAugmentedDisplay *>::const_iterator augIt = 
      augDisplays->begin();
    for (; augIt != augDisplays->end(); ++augIt) {
      GlABoundingCircle *curAug = 
	static_cast<GlABoundingCircle *> (augIt->second);
      unsigned int place = order[curAug->getHierarchyNode()];
      curAug->setPoints (g0->interpolateCenters[place],
			 g0->curInterpolation[place]);
    }
  }
}

//===============================================================
void MouseGrouseNavigate::morphGraph (GlGraphWidget *graphWidget, 
				      GraphState *g0, 
				      GraphState *g1,
				      stdext::hash_map<node, unsigned int> 
				      &order) {
  const double iterations = 150;
  Morphing morph;
  morph.init (graphWidget, g0, g1);
  float t = 0;
  for (unsigned int k = 0; k < iterations; ++k) {
    morph.interpolate (graphWidget, t);
    updateAugs (graphWidget, g0, order);
    graphWidget->draw();
    t += 1.0/iterations;
  }//end for
  graphWidget->draw();
}//end morphGraph

//===============================================================
void setupAugs (GraphState *gs,
		GlGraphWidget *graphWidget,
		stdext::hash_map<node, unsigned int> &order) {
  bool firstTime = order.empty();
  if (!firstTime) {
    gs->augPoints.resize (order.size());
    gs->augCenters.resize (order.size());
  }
  unsigned int augCounter = 0;
  const stdext::hash_map<string, GlAugmentedDisplay *> *augDisplays =
    graphWidget->getAugmentedDisplays();
  if (augDisplays != NULL) {
    stdext::hash_map<string, GlAugmentedDisplay *>::const_iterator augIt = 
      augDisplays->begin();
    for (; augIt != augDisplays->end(); ++augIt) {
      GlABoundingCircle *curAug = 
	static_cast<GlABoundingCircle *> (augIt->second);
      node hierarchyNode = curAug->getHierarchyNode();
      const vector<Coord> points = curAug->getPoints();
      if (firstTime) {
	gs->augPoints.push_back (points);
	gs->augCenters.push_back (curAug->getPos());
	order[hierarchyNode] = augCounter;
	augCounter++;
      }
      else {
	if (order.find (hierarchyNode) != order.end()) {
	  unsigned int place = order[hierarchyNode];
	  gs->augPoints[place] = points;
	  gs->augCenters[place] = curAug->getPos();
	}
	else {
	  order[hierarchyNode] = gs->augPoints.size();
	  gs->augPoints.push_back (points);
	  gs->augCenters.push_back (curAug->getPos());
	}
      }
    }
  }
}

//===============================================================
inline double checkNewZoom (double nodeCoord, double camCoord,
			    double halfCameraSceneRadius,
			    double newZoomFactor) {
  double newZoom = halfCameraSceneRadius/fabs (nodeCoord - camCoord);
  if (newZoom < newZoomFactor) return newZoom;
  else return newZoomFactor;
}

//===============================================================
void MouseGrouseNavigate::morphCamera (GlGraphWidget *graphWidget,
				       const Coord &nodePos,
				       const double nodeRadius) {
  GlGraphRenderingParameters parameters = 
      graphWidget->getRenderingParameters();
  Camera camera = parameters.getCamera();
  double newZoomFactor = camera.zoomFactor;

  newZoomFactor = checkNewZoom (nodePos.getX() + nodeRadius, 
				camera.center.getX(),
				camera.sceneRadius/2.0,
				newZoomFactor);
  newZoomFactor = checkNewZoom (nodePos.getX() - nodeRadius, 
				camera.center.getX(),
				camera.sceneRadius/2.0, 
				newZoomFactor);
  newZoomFactor = checkNewZoom (nodePos.getY() + nodeRadius, 
				camera.center.getY(),
				camera.sceneRadius/2.0,
				newZoomFactor);
  newZoomFactor = checkNewZoom (nodePos.getY() - nodeRadius, 
				camera.center.getY(),
				camera.sceneRadius/2.0, 
				newZoomFactor);
  camera.zoomFactor = newZoomFactor;
  parameters.setCamera (camera);
  graphWidget->setRenderingParameters (parameters);
}

//===============================================================
/**
 * Function to morph a node open
 */
void MouseGrouseNavigate::morphOpen (GlGraphWidget *graphWidget,
				     const node &pickedNode) {
  node hierarchyNode = openCloseHierarchy->getHierarchyNode (pickedNode);
  vector<node> nodesAdded;
  vector<edge> edgesAdded;
  vector<GlABoundingCircle *> augsAdded;
  stdext::hash_map<node, unsigned int> order;
  GraphState *g0 = new GraphState (graphWidget);
  setupAugs (g0, graphWidget, order);
  LayoutProperty *viewLayout =
    graphWidget->getGraph()->getProperty<LayoutProperty> ("viewLayout");
  const Coord curCoord = viewLayout->getNodeValue (pickedNode);
  ColorProperty *viewColor =
    graphWidget->getGraph()->getProperty<ColorProperty> ("viewColor");
  viewColor->setNodeValue (pickedNode, PROCESSING_COLOUR);
  graphWidget->draw();
  openCloseHierarchy->open (pickedNode, nodesAdded, edgesAdded, augsAdded);
  double openNodeRadius = augsAdded[0]->getRadius();
  Coord openNodePosition = augsAdded[0]->getPos();
  morphCamera (graphWidget, openNodePosition, openNodeRadius);
  openCloseHierarchy->showHideAugs (augsAdded, true);
  GraphState *g1 = new GraphState (graphWidget);
  setupAugs (g1, graphWidget, order);
  fixForAdditions (nodesAdded, edgesAdded, augsAdded, order,
		   curCoord, g0, g1, graphWidget);
  morphGraph (graphWidget, g0, g1, order);
  graphWidget->draw();
  delete g0;
  delete g1;
  metaGraphHierarchy->setListViewColour (hierarchyNode, 
					 augsAdded[0]->getColor());
}

//===============================================================
void MouseGrouseNavigate::nodesAndEdgesToRemove (Graph *graphToBeClosed,
						 vector<node> &nodesRemoved,
						 vector<edge> &edgesRemoved) {
  node curNode;
  forEach (curNode, graphToBeClosed->getNodes())
    nodesRemoved.push_back (curNode);
  edge curEdge;
  forEach (curEdge, graphToBeClosed->getEdges())
    edgesRemoved.push_back (curEdge);
}//end nodesAndEdgesToRemove

//===============================================================
void layoutReinstate (LayoutProperty *layout, GlGraphWidget *graphWidget) {
  Graph *curGraph = graphWidget->getGraph();
  LayoutProperty *curLayout = 
    curGraph->getProperty<LayoutProperty> ("viewLayout");
  node curNode;
  forEach (curNode, curGraph->getNodes())
    curLayout->setNodeValue (curNode, layout->getNodeValue (curNode));
  edge curEdge;
  forEach (curEdge, curGraph->getEdges())
    curLayout->setEdgeValue (curEdge, layout->getEdgeValue (curEdge));
}//end layoutReinstate

//===============================================================
/**
 * Function to close a node
 */
void MouseGrouseNavigate::morphClosed(GlGraphWidget *graphWidget,
				      GlABoundingCircle *pickedBoundingCircle) {
  vector<node> nodesRemoved;
  vector<edge> edgesRemoved;
  vector<GlABoundingCircle *> augsRemoved;
  openCloseHierarchy->findAugsBelow (pickedBoundingCircle, augsRemoved);
  stdext::hash_map<node, unsigned int> order;
  graphWidget->draw();
  GraphState *g0 = new GraphState (graphWidget);
  setupAugs (g0, graphWidget, order);
  Graph *graphToBeClosed = pickedBoundingCircle->getDisplaySubGraph();
  const Coord curCoord = pickedBoundingCircle->getPos();
  GraphState *g1 = new GraphState (graphWidget);
  setupAugs (g1, graphWidget, order);
  nodesAndEdgesToRemove (graphToBeClosed, nodesRemoved, edgesRemoved);
  fixForAdditions (nodesRemoved, edgesRemoved, augsRemoved, order,
		   curCoord, g1, g0, graphWidget);
  morphGraph (graphWidget, g0, g1, order);
  layoutReinstate (g0->layout, graphWidget);
  openCloseHierarchy->close (graphToBeClosed);
  openCloseHierarchy->showHideAugs (augsRemoved, false);
  graphWidget->draw();
  delete g0;
  delete g1;
}

//===============================================================
/**
 * Function to open/close nodes in the hierarchy and steer computation
 */
void MouseGrouseNavigate::openClose (const QPoint &mouseClick,
				     GlGraphWidget *graphWidget) {
  node pickedNode;
  edge pickedEdge;
  GlAugmentedDisplay *pickedAug = NULL;
  pick (mouseClick.x(), mouseClick.y(), graphWidget, true, true, true, 
	false, true, pickedNode, pickedEdge, pickedAug);

  node hierarchyNode;
  if (pickedNode.isValid ())
    hierarchyNode = openCloseHierarchy->getHierarchyNode (pickedNode);
  GlABoundingCircle *boundingCircle =
    static_cast<GlABoundingCircle *> (pickedAug);

  if (!pickedNode.isValid() && (boundingCircle == NULL)) return;
  //if node valid, it should not be behind bounding circle, open node.
  else if (hierarchyNode.isValid()) 
    morphOpen (graphWidget, pickedNode);
  else if (boundingCircle != NULL) {
    if (boundingCircle->getSelected()) {
      metaGraphHierarchy->clearSelection();
      grouseSelectionProperty->clearSelection();
    }
    morphClosed (graphWidget, boundingCircle);
  }
  else return;
}

//===============================================================
void MouseGrouseNavigate::powerOpenNodes (const node &displayNode,
					  GlGraphWidget *graphWidget) {
  queue<node> toOpen;
  toOpen.push (displayNode);
  while (!toOpen.empty()) {
    node displayNode = toOpen.front();
    toOpen.pop();
    node hierarchyNode = openCloseHierarchy->getHierarchyNode (displayNode);
    Graph *hierarchySubGraph =
      openCloseHierarchy->getHierarchyGraph (hierarchyNode);
    if (hierarchySubGraph != NULL) {
      morphOpen (graphWidget, displayNode);
      node curHierNode;
      forEach (curHierNode, hierarchySubGraph->getNodes()) {
	node curDisplayNode = openCloseHierarchy->getDisplayNode (curHierNode);
	//display nodes are not equal to hier nodes for non leafs
	if (curDisplayNode.isValid() && (curDisplayNode != curHierNode))
	  toOpen.push(curDisplayNode);
      }//end forEach
    }//end if
  }//end while
}//end powerOpen

//===============================================================
/**
 * Function to power open nodes in the hierarchy
 */

void MouseGrouseNavigate::powerOpen (const QPoint &mouseClick,
				     GlGraphWidget *graphWidget) {
  node pickedNode;
  edge pickedEdge;
  GlAugmentedDisplay *pickedAug = NULL;
  pick (mouseClick.x(), mouseClick.y(), graphWidget, true, true, true, 
	false, false, pickedNode, pickedEdge, pickedAug);
  if (pickedNode.isValid()) 
    powerOpenNodes (pickedNode, graphWidget);
}

//===============================================================
/**
 * Callback to handle mouse down events for grouse.  It also
 * listens to the keybord to see if any keys are down as well.
 * The graph widget is graph, and the Qt mouse event is e.
 */
void MouseGrouseNavigate::mPressEvent(GlGraphWidget *graph, QMouseEvent *e) {
  QMouseEvent mouseCom (QEvent::MouseButtonPress, e->pos(),
			Qt::LeftButton, e->state());
  if ((e->button() == QT_FREE_PAN_BUTTON) &&
      (e->state() == QT_FREE_PAN_KEY_DOWN))
    initStartDrag (e->pos());
  else if ((e->button() == QT_OPEN_CLOSE_BUTTON) &&
	   (e->state() == QT_OPEN_CLOSE_KEY_DOWN))
    openClose (e->pos(), graph);
  else if ((e->button() == QT_POWER_OPEN_BUTTON) &&
	   (e->state() == QT_POWER_OPEN_KEY_DOWN))
    powerOpen (e->pos(), graph);
  else if ((e->button() == QT_SELECT_BUTTON) &&
	   (e->state() == QT_SELECT_KEY_DOWN))
    select (e->pos(), graph);
  else e->ignore();
}

//===============================================================
/**
 * Callback to handle mouse up events for grouse.  It also
 * listens to the keybord to see if any keys are down as well.
 * The graph widget is graph, and the Qt mouse event is e.
 */
void MouseGrouseNavigate::mReleaseEvent(GlGraphWidget *graph, QMouseEvent *e) {}

//===============================================================
/**
 * Callback to handle when the mouse is moving.  It will look also
 * to see if any keys or mouse buttons are pressed.  The graph widget
 * is graph, and the mouse event is e.
 */
void MouseGrouseNavigate::mMoveEvent(GlGraphWidget *graph, QMouseEvent *e) {
  QMouseEvent mouseCom (QEvent::MouseMove, e->pos(),
			Qt::LeftButton, e->state());
  if (e->state() == QT_FREE_PAN_DRAG_STATE) {
    pan (e->pos(), graph);
    graph->draw();
  }
  else e->ignore();
}

//===============================================================
/**
 * Callback to handle when a key on the keybord is pressed.  The
 * graph widget is g and the QKeyEvent is e.
 */
void MouseGrouseNavigate::keyPressEvent(GlGraphWidget *graph, QKeyEvent *e) {
  int delta = (e->isAutoRepeat() ? 3 : 1);
  e->accept();
  switch(e->key()) {
    case Qt::Key_Right: graph->translateCamera(delta * 2,0,0); break;
    case Qt::Key_Left: graph->translateCamera(-1 * delta * 2,0,0); break;
    case Qt::Key_Down: graph->translateCamera(0,-1 * delta * 2,0); break;
    case Qt::Key_Up: graph->translateCamera(0,delta * 2,0); break;
    case Qt::Key_Prior: graph->zoom(delta); break;
    case Qt::Key_Next: graph->zoom(-1 * delta); break;
    case Qt::Key_Home: graph->translateCamera(0,0,-1 * delta * 2); break;
    case Qt::Key_End: graph->translateCamera(0,0,delta * 2); break;
    case Qt::Key_Insert: graph->rotateScene(0,0,-1 * delta * 2); break;
    case Qt::Key_Delete : graph->rotateScene(0,0,delta * 2); break;
    default: 
      e->ignore();
      return;
  }
  graph->draw();
}
  
//===============================================================
/**
 * Callback to handle when a key on the keybord is pressed.  The
 * graph widget is g and the QKeyEvent is e.
 */
void MouseGrouseNavigate::keyReleaseEvent(GlGraphWidget *g, QKeyEvent *e) {}

//===============================================================
/**
 * Callback to handle when the mouse wheel is used.  The graph
 * widget is g and the wheel event is e.
 */
// according to Qt's doc, this constant has been defined by wheel mouse vendors
// we need it to interpret the value of QWheelEvent->delta()
#define WHEEL_DELTA (120)
void MouseGrouseNavigate::wheelEvent(GlGraphWidget *graph, QWheelEvent *e) {
  if (e->orientation() == Qt::Vertical) {
    graph->zoomXY(e->delta() / WHEEL_DELTA, e->x(), e->y());
    graph->draw();
  }
  else {
    e->ignore();
  }
}

//===============================================================
/**
 * Event filter for all the mouse events on the grouse view window.
 * Determines which private navigation function to call.
 */
bool MouseGrouseNavigate::eventFilter (QObject *widget, QEvent *e) {
  GlGraphWidget *glGraphWidget = static_cast<GlGraphWidget *> (widget);
  if (e->type() == QEvent::MouseButtonPress) 
    mPressEvent(glGraphWidget, static_cast<QMouseEvent *> (e));
  else if (e->type() == QEvent::MouseButtonRelease)
    mReleaseEvent (glGraphWidget, static_cast<QMouseEvent *> (e));
  else if (e->type() == QEvent::MouseMove)
    mMoveEvent (glGraphWidget, static_cast<QMouseEvent *> (e));
  else if (e->type() == QEvent::KeyPress)
    keyPressEvent (glGraphWidget, static_cast<QKeyEvent *> (e));
  else if (e->type() == QEvent::KeyRelease)
    keyPressEvent (glGraphWidget, static_cast<QKeyEvent *> (e));
  else if (e->type() == QEvent::Wheel)
    wheelEvent (glGraphWidget, static_cast<QWheelEvent *> (e));
  else return false;
  return true;
}
