/**
 * Implements data structures and functions to open an close nodes
 * in the meta-graph hierarchy.
 */
#ifndef __OPEN_CLOSE_H__
#define __OPEN_CLOSE_H__

#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <string>
#include <vector>
#include <tulip/TlpTools.h>
#include <tulip/Graph.h>
#include <tulip/GraphProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/ForEach.h>

/**
 * This class provides operations to open and close nodes and
 * meta edges.
 */
class OpenClose {
 private:
  tlp::Graph *closedSubgraphs;
  tlp::GraphProperty *groupedMetaGraphs;
  const std::string GROUP_SUB_GRAPH_PROPERTY;
  const std::string OPEN_CLOSE_NAME;

  /**
   * Loads all nodes in the src subgraph into the dest subgraph
   */
  inline void copy (tlp::Graph *src, tlp::Graph *dest) {
    tlp::node curNode;
    forEach (curNode, src->getNodes()) dest->addNode (curNode);
    tlp::edge curEdge;
    forEach (curEdge, src->getEdges()) dest->addEdge (curEdge);
  }

 public:
  inline OpenClose (tlp::Graph *displayGraph) :
    GROUP_SUB_GRAPH_PROPERTY("groupSubGraphMetaGraphs"),
    OPEN_CLOSE_NAME("closed subgraphs") {
    closedSubgraphs = displayGraph->getRoot()->addSubGraph();
    copy (displayGraph, closedSubgraphs);
    closedSubgraphs->setAttribute<std::string> ("name", OPEN_CLOSE_NAME);
    groupedMetaGraphs = 
      displayGraph->getRoot()
      ->getLocalProperty<tlp::GraphProperty>(GROUP_SUB_GRAPH_PROPERTY);
    groupedMetaGraphs->setAllNodeValue (NULL);
  }

  inline ~OpenClose () {
    tlp::Graph *root = closedSubgraphs->getRoot();
    root->delAllSubGraphs (closedSubgraphs);
  }

  /**
   * Function to return the grouped meta graph of a node.  If no
   * such group exists, it returns null.
   * \param metaNode The node for which we want to find the grouped
   * meta graph.
   */
  inline tlp::Graph *getGroupedGraph (const tlp::node &metaNode) {
    if (!closedSubgraphs->isElement (metaNode)) return NULL;
    return groupedMetaGraphs->getNodeValue (metaNode);
  }

  /**
   * Function to return the root of where the grouped graphs are stored.
   */
  inline tlp::Graph *getGroupRoot () {
    return closedSubgraphs;
  }

  /**
   * Function to close a subgraph into a meta-node.
   * \param displayGraph The graph displayed which will be grouped.
   * \param subgraph The subgraph to be grouped.
   * \param pos The position of the node after grouping
   * \param shape The shape of the node after closing.
   * \param color The colour of the node after closing.
   * \param borderWidth The border width after closing.
   * \param size The size of the node after closing.
   * \param name The label of the node after closing
   */
  tlp::node close (tlp::Graph *displayGraph, tlp::Graph *subgraph,
		   const tlp::Coord &pos,
		   const int shape, const tlp::Color &color, 
		   const double borderWidth,
		   const tlp::Size &size,
		   const std::string &label);

  /**
   * Function to open a meta-node.
   * \param displayGraph The graph displayed which will be ungrouped.
   * \param metaNode The meta-node to be ungrouped.
   * \param subgraphName The name of the subgraph to be grouped.
   */
  void open (tlp::Graph *displayGraph, tlp::node metaNode);
};

#endif

