#include <tulip/ForEach.h>
#include <tulip/StableIterator.h>
#include <tulip/DoubleProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/ColorProperty.h>
#include "OpenCloseHierarchy.h"

using namespace std;
using namespace tlp;
const Coord noPos (0, 0, 0);
const Size noSize (1, 1, 1);
const node invalidNode (NO_PARENT);

//================================================================
/**
 * The function initializes the rendering widgets to their initial
 * startup state after a file has been loaded
 */
inline void copy (Graph *src, Graph *dest) {
  node curNode;
  forEach (curNode, src->getNodes()) dest->addNode (curNode);
  edge curEdge;
  forEach (curEdge, src->getEdges()) dest->addEdge (curEdge);
}

OpenCloseHierarchy::OpenCloseHierarchy (Graph *hierarchy, 
					const node &hierarchyRootNode,
					Graph *inputGraph,
					MetaNodeBoundingCircles *boundingCircles, bool scaleMode) :
  hierarchy(hierarchy), inputGraph(inputGraph),
  boundingCircles(boundingCircles), noLayout (true), scaleMode(scaleMode) {
  openCloseHierarchyGraph = inputGraph->getRoot()->addSubGraph();
  copy (inputGraph, openCloseHierarchyGraph);

  //initialize the tables
  displaySubIDToHierNode.setAll (invalidNode);
  hierNodeToDisplayMetaNode.setAll (invalidNode);

  //layout for the hierarchy
  displayLayout = openCloseHierarchyGraph
    ->getLocalProperty<LayoutProperty> ("viewLayout");
  displaySizes = openCloseHierarchyGraph
    ->getLocalProperty<SizeProperty> ("viewSize");
  hierarchyMetaGraphs = 
    hierarchy->getProperty<GraphProperty> ("viewMetaGraph");
  Graph *openCloseSubGraph = tlp::newCloneSubGraph (openCloseHierarchyGraph);

  //copy the original input sizes to the new local sizes
  SizeProperty *inputSizes = 
    inputGraph->getProperty<SizeProperty> ("viewSize");
  node inputNode;
  forEach (inputNode, inputGraph->getNodes()) {
    Size inputSize = inputSizes->getNodeValue (inputNode);
    displaySizes->setNodeValue (inputNode, inputSize);
  }//end forEach
  
  //add an node to represent the top level of the hierarchy and initialize
  //the hierarchy
  initializeOpenCloseHierarchy (hierarchy, openCloseSubGraph, 
				hierarchyRootNode);
  openClose = new OpenClose (openCloseHierarchyGraph);
  if (noLayout) {
    Iterator<Graph *> *subgraphs = openCloseSubGraph->getSubGraphs();
    while (subgraphs->hasNext()) {
      Graph *subGraph = subgraphs->next();
      closeAll (subGraph);
    }//end while
    delete subgraphs;
    drawSteeredNode (hierarchy, NULL, openCloseSubGraph);
    copyLayoutHierarchyToGroup (hierarchy, openCloseSubGraph);
  }//end if
}//end OpenCloseHierarchy

//================================================================
void OpenCloseHierarchy::deleteEdgeOrderAttributeTables (Graph *curHierarchy) {
  EdgeOrderTable *curTable = 
    curHierarchy->getAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE);
  if (curTable != NULL) delete curTable;
  node curHierNode;
  forEach (curHierNode, curHierarchy->getNodes()) {
    Graph *subGraph = hierarchyMetaGraphs->getNodeValue (curHierNode);
    if (subGraph != NULL)
      deleteEdgeOrderAttributeTables (subGraph);
  }
}

//================================================================
OpenCloseHierarchy::~OpenCloseHierarchy () {
  delete openClose;
  deleteEdgeOrderAttributeTables (hierarchy);
}//end ~OpenCloseHierarchy

//================================================================
Color getBaseColour (Graph *graphToColour) {
  if (isType (graphToColour, TREE_IDENT))
    return TREE_COLOUR;
  else if (isType (graphToColour, BICONNECTED_IDENT))
    return BICONNECTED_COLOUR;
  else if (isType (graphToColour, EMBEDDER_IDENT))
    return EMBEDDER_COLOUR;
  else if (isType (graphToColour, STRENGTH_IDENT))
    return STRENGTH_COLOUR;
  else if (isType (graphToColour, NEAR_COMPLETE_IDENT) ||
	   isType (graphToColour, OLD_NEAR_COMPLETE_IDENT))
    return COMPLETE_COLOUR;
  else if (isType (graphToColour, CONNECTED_IDENT))
    return CONNECTED_COLOUR;
  else 
    return UNKNOWN_COLOUR;
}//end getBaseColor

//================================================================
void OpenCloseHierarchy::colourMyGraph (Graph *graphToColour) {
  ColorProperty *graphColours = 
    graphToColour->getProperty<ColorProperty> ("viewColor");
  ColorProperty *graphBorderColours =
    graphToColour->getProperty<ColorProperty> ("viewBorderColor");
  DoubleProperty *graphBorderWidths =
    graphToColour->getRoot()->getProperty<DoubleProperty> ("viewBorderWidth");

  node curNode;
  forEach (curNode, graphToColour->getNodes()) {
    Graph *curSubGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    if (curSubGraph != NULL) {
      Color baseColour = getBaseColour (curSubGraph);
      Color nodeColour = baseColour;
      nodeColour.setA (NODE_ALPHA);
      graphColours->setNodeValue (curNode, nodeColour);
      graphBorderColours->setNodeValue (curNode, META_NODE_BORDER_COLOUR);
      graphBorderWidths->setNodeValue (curNode, META_NODE_BORDER_WIDTH);
    }
  }//end forEach
}//end colourMyGraph

//================================================================
node makeGlyph (Graph *displaySubGraph, Graph *hierarchySubGraph) {
  StableIterator<edge> delEdges (displaySubGraph->getEdges());
  while (delEdges.hasNext())
    displaySubGraph->delAllEdge(delEdges.next());
  node centre = displaySubGraph->addNode();
  node curNode;
  forEach (curNode, displaySubGraph->getNodes())
    if (curNode != centre) displaySubGraph->addEdge (centre, curNode);
  hierarchySubGraph->setAttribute<node> ("root", centre); 
  return centre;
}

//================================================================
void
OpenCloseHierarchy::initializeOpenCloseHierarchy (Graph *curHierarchy,
						  Graph *curDisplayGraph,
						  const node &hierarchyMetaNode) {
  ColorProperty *openCloseColour = 
    openCloseHierarchyGraph->getProperty<ColorProperty> ("viewColor");
  string name = curHierarchy->getAttribute<string> ("name");
  curDisplayGraph->setAttribute<string> ("name", name);
  displaySubIDToHierNode.set (curDisplayGraph->getId(), hierarchyMetaNode);
  curHierarchy->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE, NULL);
  curHierarchy->setAttribute<bool> (HAS_LAYOUT_ATTRIBUTE, !noLayout);
  colourMyGraph (curHierarchy);

  //set the input layout to zero if none provided
  LayoutProperty *inputLayout = 
    inputGraph->getProperty<LayoutProperty> ("viewLayout");
  if (noLayout) {
    displayLayout->setAllNodeValue (noPos);
    LayoutProperty *curHierarchyLayout =
      curHierarchy->getLocalProperty<LayoutProperty> ("viewLayout");
    curHierarchyLayout->setAllNodeValue (noPos);
    DoubleProperty *curHierarchyRotations =
      curHierarchy->getLocalProperty<DoubleProperty> ("viewRotation");
    curHierarchyRotations->setAllNodeValue (0);
  }//end if

  node curNode;
  forEach (curNode, curHierarchy->getNodes()) {
    Graph *nodeMetaGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    //add only nodes in the original graph
    if (nodeMetaGraph == NULL) {
      if (!inputGraph->isElement (curNode)) continue;
      curDisplayGraph->addNode (curNode);
      whereNodeBelongs[curNode] = curDisplayGraph;
      openCloseColour->setNodeValue (curNode, LEAF_NODE_COLOUR);
      if (!noLayout)
	displayLayout->setNodeValue (curNode, 
				       inputLayout->getNodeValue (curNode));
    }//end if
    //recursively go through the hierarchy
    else
      initializeOpenCloseHierarchy (nodeMetaGraph,
				    curDisplayGraph->addSubGraph(),
				    curNode);
  }//end forEach

  forEach (curNode, curDisplayGraph->getNodes()) {
    edge curEdge;
    forEach (curEdge, inputGraph->getOutEdges (curNode)) {
      node opposite = inputGraph->opposite (curEdge, curNode);
      if ((curDisplayGraph->isElement (opposite)) &&
	  (whereEdgeBelongs.find (curEdge) == whereEdgeBelongs.end())) {
	whereEdgeBelongs[curEdge] = curDisplayGraph;
	curDisplayGraph->addEdge (curEdge);
	if (!noLayout)
	  displayLayout->setEdgeValue (curEdge,
				       inputLayout->getEdgeValue (curEdge));
	Color edgeColour = EDGE_COLOUR;
	edgeColour.setA (EDGE_ALPHA);
	openCloseColour->setEdgeValue (curEdge, edgeColour);
      }//end if
    }//end forEachEdge
  }//end forEach nodes
  
  if (isType (curHierarchy, NEAR_COMPLETE_IDENT)) {
    SizeProperty *displaySize = 
      curDisplayGraph->getProperty<SizeProperty> ("viewSize");
    ColorProperty *displayColour = 
      curDisplayGraph->getProperty<ColorProperty> ("viewColor");
    IntegerProperty *displayShape = 
      curDisplayGraph->getProperty<IntegerProperty> ("viewShape");
    curHierarchy->setAttribute<bool> ("isComplete", true);
    edge curEdge;
    forEach (curEdge, curDisplayGraph->getEdges())
      whereEdgeBelongs.erase (curEdge);
    node root = makeGlyph (curDisplayGraph, curHierarchy);
    whereNodeBelongs[root] = curDisplayGraph;
    displaySize->setNodeValue (root, Size (2,2,0));
    displayShape->setNodeValue (root, GLYPH_CENTRE_SHAPE);
    displayColour->setNodeValue (root, GLYPH_EDGE_COLOUR);
    Color edgeColour = GLYPH_EDGE_COLOUR;
    edgeColour.setA (150);
    forEach (curEdge, curDisplayGraph->getEdges()) {
      whereEdgeBelongs[curEdge] = curDisplayGraph;
      displayColour->setEdgeValue (curEdge, edgeColour);
    }
  }//end if
}//end initialize hierarchy

//================================================================
void OpenCloseHierarchy::closeAll (Graph *curGraph) {
  if ((curGraph->numberOfEdges() == 0) && (curGraph->numberOfNodes() == 0))
    return;
  double numberOfNodes = curGraph->numberOfNodes();
  Iterator<Graph *> *curGraphSubGraphs = curGraph->getSubGraphs ();
  while (curGraphSubGraphs->hasNext()) {
    Graph *curGraphSubGraph = curGraphSubGraphs->next();
    closeAll (curGraphSubGraph);
  } delete curGraphSubGraphs;

  Graph *father = curGraph->getSuperGraph();
  node fatherHierarchyNode = displaySubIDToHierNode.get (father->getId());
  Graph *containingHierarchyGraph = 
    hierarchyMetaGraphs->getNodeValue (fatherHierarchyNode);

  SizeProperty *hierarchySizes =
    containingHierarchyGraph->getProperty<SizeProperty> ("viewSize");
  SizeProperty *hierarchyBoundingBox =
    containingHierarchyGraph->getLocalProperty<SizeProperty> ("boundingBox");
  double theSize = sqrt (numberOfNodes);
  Size nodeSize (theSize, theSize, 1.0);
  Size boundingBox (M_SQRT2*theSize, M_SQRT2*theSize);
  node hierarchyNode = displaySubIDToHierNode.get (curGraph->getId());
  hierarchySizes->setNodeValue (hierarchyNode, nodeSize);
  hierarchyBoundingBox->setNodeValue (hierarchyNode, boundingBox);
  close (curGraph);
}//end closeAllAndUnsetLayout

//================================================================
Graph *OpenCloseHierarchy::findWhereEdgeBelongs (const node &metaNode,
						 const node &opposite) {
  Graph *metaSuper = whereNodeBelongs[metaNode]->getSuperGraph();
  if (metaSuper->isElement (opposite)) return metaSuper;
  else {
    Graph *super = whereNodeBelongs[opposite];
    while (!super->isElement (metaNode)) super = super->getSuperGraph();
    return super;
  }//end else
}//end findWhereEdgeBelongs

//================================================================
void OpenCloseHierarchy::close (Graph *closeHierarchySub) {  
  node hierarchyNode = displaySubIDToHierNode.get (closeHierarchySub->getId());
  if (!hierarchyNode.isValid()) {
    cerr << "The graph " << closeHierarchySub->getAttribute<string> ("name")
	 << " does not exist in the in the display graph." << endl;
    return;
  }

  Graph *father = closeHierarchySub->getSuperGraph();
  node fatherHierarchyNode = displaySubIDToHierNode.get (father->getId());
  Graph *hierarchyFather = 
    hierarchyMetaGraphs->getNodeValue (fatherHierarchyNode);
  LayoutProperty *hierarchyLayout = 
    hierarchyFather->getProperty<LayoutProperty> ("viewLayout");
  Coord hierPos = hierarchyLayout->getNodeValue (hierarchyNode);
  ColorProperty *hierarchyColours =
    hierarchyFather->getProperty<ColorProperty> ("viewColor");
  Color hierColor = hierarchyColours->getNodeValue (hierarchyNode);
  SizeProperty *hierarchySizes =
    hierarchyFather->getProperty<SizeProperty> ("viewSize");
  string label = closeHierarchySub->getAttribute<string> ("name");
  Size hierSize = hierarchySizes->getNodeValue (hierarchyNode);

  node addedNode = openClose->close (openCloseHierarchyGraph, 
				     closeHierarchySub, 
				     hierPos, META_NODE_SHAPE, 
				     hierColor, META_NODE_BORDER_WIDTH,
				     hierSize, label);
  hierNodeToDisplayMetaNode.set (hierarchyNode.id, addedNode);
  closeHierarchySub->addNode (addedNode);
  whereNodeBelongs[addedNode] = closeHierarchySub;

  //place meta edges in the subgraphs where they belong
  Graph *groupRoot = openClose->getGroupRoot ();
  groupRoot->addNode (addedNode);
  StableIterator<edge> 
    inOutEdges (openCloseHierarchyGraph->getInOutEdges (addedNode));
  while (inOutEdges.hasNext()) {
    edge curEdge = inOutEdges.next();
    node opposite = openCloseHierarchyGraph->opposite (curEdge, addedNode);
    Graph *curEdgeBelongs = findWhereEdgeBelongs (addedNode, opposite);
    curEdgeBelongs->addEdge (curEdge);
    whereEdgeBelongs[curEdge] = curEdgeBelongs;
    groupRoot->addEdge (curEdge);
  }//end forEach
}//end close

//================================================================
node OpenCloseHierarchy::getHierarchyNode (const node &groupNode) {
  if (!groupNode.isValid()) return invalidNode;
  node hierarchyNode = invalidNode;
  Graph *curGroupNodeGraph = openClose->getGroupedGraph (groupNode);
  if (curGroupNodeGraph != NULL) {
    Graph *displayHierarchyGraph = whereNodeBelongs[groupNode];
    hierarchyNode = 
      displaySubIDToHierNode.get (displayHierarchyGraph->getId());
  }//end if
  else if (inputGraph->isElement (groupNode)) 
    hierarchyNode = groupNode; //leaves are the same in both graphs

  return hierarchyNode;
}//end getHierarchyNode

//================================================================
node OpenCloseHierarchy::getDisplayNode (const node &hierarchyNode) {
  if (!hierarchyNode.isValid()) return invalidNode;
  node displayNode = hierNodeToDisplayMetaNode.get (hierarchyNode.id);
  if (inputGraph->isElement (hierarchyNode))
    displayNode = hierarchyNode;
  else if (displayNode.isValid() && 
	   !openCloseHierarchyGraph->isElement (displayNode))
    displayNode = invalidNode;
  return displayNode;
}//end getDisplayNode

//================================================================
GlABoundingCircle *OpenCloseHierarchy::
getBoundingCircle (const node &hierarchyNode) {
  return boundingCircles->getBoundingCircle (hierarchyNode);
}//end getBoundingCircle

//================================================================
Graph *OpenCloseHierarchy::getHierarchyGraph (const node &hierarchyNode) {
  return hierarchyMetaGraphs->getNodeValue (hierarchyNode);
}//end getHierarchyGraph

//================================================================
void OpenCloseHierarchy::copyLayoutHierarchyToGroup (Graph *hierarchyGraph,
						     Graph *groupGraph) {
  LayoutProperty *hierarchyProperty =
    hierarchyGraph->getProperty<LayoutProperty> ("viewLayout");
  LayoutProperty *groupProperty =
    groupGraph->getProperty<LayoutProperty> ("viewLayout");
  node groupNode;
  forEach (groupNode, groupGraph->getNodes()) {
    node hierarchyNode = getHierarchyNode (groupNode);
    groupProperty->setNodeValue (groupNode,
				 hierarchyProperty
				 ->getNodeValue (hierarchyNode));
    edge adjEdge;
    forEach (adjEdge, groupGraph->getInOutEdges (groupNode)) {
      node source = groupGraph->source (adjEdge);
      node target = groupGraph->target (adjEdge);
      node hierarchySource = getHierarchyNode (source);
      node hierarchyTarget = getHierarchyNode (target);
      if (hierarchySource.isValid() && hierarchyTarget.isValid()) {
	edge hierarchyEdge = 
	  hierarchyGraph->existEdge (hierarchySource, hierarchyTarget);
	groupProperty->setEdgeValue (adjEdge,
				     hierarchyProperty
				     ->getEdgeValue (hierarchyEdge));
      }//end if
    }//end for
  }//end for
}//end copyLayoutHierarchyToOpenClose

//================================================================
void OpenCloseHierarchy::showHideAugs (vector<GlABoundingCircle *> &augs,
				       bool show) {
  vector<GlABoundingCircle *>::iterator augsIt = augs.begin();
  for (; augsIt != augs.end(); ++augsIt) {
    if (show)
      boundingCircles->showBoundingCircle ((*augsIt)->getHierarchyNode());
    else
      boundingCircles->hideBoundingCircle ((*augsIt)->getHierarchyNode());
  }//end for
}//end showHideAugs

//================================================================
void OpenCloseHierarchy::findAugsBelow (GlABoundingCircle *curAug,
					std::vector<GlABoundingCircle *> 
					&augsBelow) {
  augsBelow.push_back (curAug);
  node hierarchyNode = curAug->getHierarchyNode();
  Graph *subGraph = hierarchyMetaGraphs->getNodeValue (hierarchyNode);
  node curSubNode;
  forEach (curSubNode, subGraph->getNodes()) {
    GlABoundingCircle *subAug = 
      boundingCircles->getBoundingCircle (curSubNode);
    if (subAug != NULL) {
      Graph *displaySubGraph = subAug->getDisplaySubGraph();
      if (displaySubGraph->numberOfNodes() > 1)
	findAugsBelow (subAug, augsBelow);
    }//end if
  }//end forEach 
}

//================================================================
void OpenCloseHierarchy::open (node metaNode,
			       vector<node> &nodesAdded,
			       vector<edge> &edgesAdded,
			       vector<GlABoundingCircle *> &augsAdded) {
  if (!openCloseHierarchyGraph->isElement (metaNode)) {
    std::cerr << "The node " << metaNode.id 
	      << " does not exist in the in the display graph." << std::endl;
    return;
  }
  
  Graph *groupGraph = openClose->getGroupedGraph (metaNode);
  if (groupGraph == NULL) return;
  Graph *groupRoot = openClose->getGroupRoot ();

  //remove strutures for the metanode
  Graph *metaNodeBelongs = whereNodeBelongs[metaNode];
  edge curEdge;
  forEach (curEdge, openCloseHierarchyGraph->getInOutEdges (metaNode))
    whereEdgeBelongs.erase (curEdge);
  whereNodeBelongs.erase (metaNode);
  node hierarchyNode = displaySubIDToHierNode.get (metaNodeBelongs->getId());
  hierNodeToDisplayMetaNode.set (hierarchyNode.id, invalidNode);

  //layout the meta-node if required
  Graph *metaNodeHierarchyGraph = 
    hierarchyMetaGraphs->getNodeValue(hierarchyNode);
  bool hasLayout = 
    metaNodeHierarchyGraph->getAttribute<bool> (HAS_LAYOUT_ATTRIBUTE);
  if (!hasLayout) {
    node fatherMetaNode = 
      displaySubIDToHierNode.get (metaNodeBelongs->getSuperGraph()->getId());
    Graph *hierarchyFather = 
      hierarchyMetaGraphs->getNodeValue (fatherMetaNode);
    ColorProperty *hierarchyColours =
      hierarchyFather->getProperty<ColorProperty> ("viewColor");
    Color hierColor = hierarchyColours->getNodeValue (hierarchyNode);
    boundingCircles->addBoundingCircle (openCloseHierarchyGraph,
					metaNodeBelongs,
					hierarchyNode, noPos, 
					noSize, hierColor);
    augsAdded.push_back (boundingCircles->getBoundingCircle (hierarchyNode));
    drawSteeredNode (metaNodeHierarchyGraph, hierarchyFather, metaNodeBelongs);
    copyLayoutHierarchyToGroup (metaNodeHierarchyGraph, groupGraph);
  }//end if no layout
 
  //replace all nodes and edges in the meta node graph
  node curNode;
  unsigned int numberOfNodes = 0;
  node groupGraphNodes[groupGraph->numberOfNodes()];
  forEach (curNode, groupGraph->getNodes()) {
    whereNodeBelongs[curNode]->addNode (curNode);
    nodesAdded.push_back (curNode);
    groupGraphNodes[numberOfNodes] = curNode;
    ++numberOfNodes;
  }
  forEach (curEdge, groupGraph->getEdges()) {
    whereEdgeBelongs[curEdge]->addEdge (curEdge);
    edgesAdded.push_back (curEdge);
  }
  openClose->open (openCloseHierarchyGraph, metaNode);
  if (!hasLayout) {
    if (!scaleMode) {
      cleanupDisplay (metaNodeBelongs);
    }
    else
      scaleDown (metaNodeBelongs);
  }
  else {
    findAugsBelow (boundingCircles->getBoundingCircle (hierarchyNode),
		   augsAdded);
    updateBoundingCircles (hierarchy, noPos, 0, 1.0);
  }

  //replace all edges if their end nodes are in the display
  for (unsigned int i = 0; i < numberOfNodes; ++i) {
    edge curEdge;
    forEach (curEdge, groupRoot->getInOutEdges (groupGraphNodes[i])) {
      node opposite = groupRoot->opposite (curEdge, groupGraphNodes[i]);
      if (openCloseHierarchyGraph->isElement (opposite)) {
	if ((whereEdgeBelongs.find (curEdge) != whereEdgeBelongs.end()) &&
	    (whereEdgeBelongs[curEdge]->isElement (curEdge))) 
	  continue;
	else if ((whereEdgeBelongs.find (curEdge) != whereEdgeBelongs.end()) &&
		 (!whereEdgeBelongs[curEdge]->isElement (curEdge)))
	  whereEdgeBelongs[curEdge]->addEdge (curEdge);
	else {
	  Graph *curEdgeBelongs = 
	    findWhereEdgeBelongs (groupGraphNodes[i], opposite);
	  edge added = curEdgeBelongs->addEdge (groupGraphNodes[i], opposite);
	  whereEdgeBelongs[added] = curEdgeBelongs;
	}//end else
      }//end if
    }//end for
  }//end for nodes
}//end open
