
/**
 * Keeps track of a subgraph hierarchy for opening and closing nodes.
 */
#ifndef __OPEN_CLOSE_HIERARCHY_H__
#define __OPEN_CLOSE_HIERARCHY_H__
#include <string>
#if (__GNUC__ < 3)
#  include <hash_map>
#else
#  include <ext/hash_map>
#endif
#include <tulip/MutableContainer.h>
#include <tulip/TlpTools.h>
#include <tulip/ForEach.h>
#include <tulip/Graph.h>
#include <tulip/GlGraph.h>
#include <tulip/LayoutProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/GraphProperty.h>
#include <tulip/MutableContainer.h>
#include "MetaNodeBoundingCircle.h"
#include "ColourShapeDefaults.h"
#include "OpenClose.h"
#include "SteeredLayout.h"
#include "HierarchyMap.h"

#define OPEN_CLOSE_HIERARCHY_NAME ("display graph")

class OpenCloseHierarchy {
 private:
  bool noLayout;
  bool scaleMode;
  tlp::Graph *openCloseHierarchyGraph; //the display graph of the hierarchy
  tlp::Graph *inputGraph; //pointer to orignal input graph
  tlp::Graph *hierarchy; //pointer to meta-graph hierarchy
 
  OpenClose *openClose;
  tlp::LayoutProperty *displayLayout; //hierarchy node positions
  tlp::SizeProperty *displaySizes; //hierarchy node sizes
  tlp::GraphProperty *hierarchyMetaGraphs; //property for hierarchy meta graphs

  MetaNodeBoundingCircles *boundingCircles;
  tlp::MutableContainer<tlp::node> displaySubIDToHierNode;
  tlp::MutableContainer<tlp::node> hierNodeToDisplayMetaNode;
  stdext::hash_map<tlp::node, tlp::Graph *> whereNodeBelongs;
  stdext::hash_map<tlp::edge, tlp::Graph *> whereEdgeBelongs;

  /**
   * A function to set up the subgraphs of the open close hierarchy.
   * \param hierarchy the graph hierarchy
   * \param curOpenCloseHierarchy the current display graph.
   * \param containingNode The node which contains the current hierarchy
   */
  void initializeOpenCloseHierarchy (Graph *hierarchy, 
				     Graph *curOpenCloseHierarchy,
				     const node &containingNode);

  /**
   * Private function to delete attribute tables.  Used by destructor only.
   * \param curHierarchy A subgraph of the graph hierarchy.
   */
  void deleteEdgeOrderAttributeTables (Graph *curHierarchy);

  /**
   * A function to colour the graph by topology type.
   * \param graphToColour The graph which we would like to colour
   */
  void colourMyGraph (Graph *graphToColour);

  /**
   * Given two nodes adjacent to an edge, it finds where in the subgraph
   * hierarchy those two nodes belong.
   * \param metaNode the meta node end of the edge
   * \param opposite the other end of the edge
   */
  Graph *findWhereEdgeBelongs (const node &metaNode, const node &opposite);

  /**
   * Function to copy a hierarchy subgraph layout to a group/display
   * subgraph layout.
   * \param hierarchyGraph the hierarchy sub graph
   * \param groupGraph the group/display sub graph
   */
  void copyLayoutHierarchyToGroup (Graph *hierarchyGraph,
				   Graph *groupGraph);

  /**
   * Function to update the bounding circle positions, size, and rotation.
   * \param curHierGraph name of the hierarchy meta graph
   * \param pos New position of the sub graph
   * \param rotation New rotation of the subgraph
   */
  void updateBoundingCircles (Graph *curHierGraph, 
			      const tlp::Coord &pos, 
			      double rotation, float scale);

  /**
   * Function to update the layouts along a particular chain.
   * \param metaGraph The graph which has been opened
   * \param cutDisplayChain The chain of graphs leading to the one
   * which has been opened
   */
  void redrawChain (Graph *metaGraph,
		    std::vector<tlp::Graph *> &cutDisplayChain);

  /**
   * Function to update the positions of the nodes along the cut.
   * \param cutDisplayChain The chain of graphs leading to the one
   * which has been opened
   */
  void updateCutLayout (std::vector<tlp::Graph *> &cutDisplayChain);

  void updateCutLayout (Graph *curHierarchySubGraph,
			const Coord &cumulativeTrans,
			double cumulativeRotation,
			float cumulativeScale);

  void scaleDown (tlp::Graph *metaGraph);
  

 public:
  OpenCloseHierarchy (Graph *hierarchy, const tlp::node &hierarchyRoot,
		      Graph *inputGraph,
		      MetaNodeBoundingCircles *boundingCircles,
		      bool scaleMode);
  
  ~OpenCloseHierarchy (); 

  inline void setScaleMode (bool newScaleMode) { scaleMode = newScaleMode; }

  /**
   * Returns a subgraph pointer to the hierarchy
   */
  inline Graph *getOpenCloseHierarchyGraph () {
    return openCloseHierarchyGraph;
  }//end getOpenCloseHierarchy

  /**
   * Function to get the hierarchy node given a group/display node.  Returns
   * the all closed node.
   * \param groupNode the display node. 
   */
  tlp::node getHierarchyNode (const tlp::node &displayNode);

  /**
   * Function to get the hierarchy subgraph given the hierarchy node.
   * \param hierarchyNode The hierarchy node.
   */
  tlp::Graph *getHierarchyGraph (const tlp::node &hierarchyNode);

  /**
   * Function to get a display node given a hierarchy node.
   * \param hierarchyNode The hierarchy node.
   */
  tlp::node getDisplayNode (const tlp::node &hierarchyNode);

  /**
   * Function to get a bouding circle given a hierarchy node
   * \param hierarchyNode The hierarchy node.
   */
  GlABoundingCircle *getBoundingCircle (const tlp::node &hierarchyNode);

  /**
   * Provides a list of all the bounding circles below this
   * bounding circle that are open.
   * \param curAug The current augmented display
   * \param augsBelow The list of augmented displays below
   */
  void findAugsBelow (GlABoundingCircle *curAug,
		      std::vector<GlABoundingCircle *> &augsBelow);

  /**
   * Function to show or hide a list of augmented displays
   * \param augs The list of augmented displays
   * \param show Boolean indicating if we are showing or hiding
   */
  void showHideAugs (std::vector<GlABoundingCircle *> &augs,
		     bool show);

  /**
   * Function to close a subgraph into a meta-node.
   * \param graphToBeClosed The graph to be closed in the openCloseHierarchy
   */
  void close (tlp::Graph *graphToBeClosed);

  /**
   * Function to close all meta nodes in the hierarchy and set them
   * to an unlayout value.
   * \param curGraph The graph of the openCloseHierarchy where closing
   * begins
   */
  void closeAll (tlp::Graph *curGraph);

  /**
   * Function to open a meta-node in the hierarchy.
   * \param metaNode The meta-node to be ungrouped.
   * \param nodesAdded The nodes added due to the open event
   * \param edgesAdded The edges added due to the open event
   * \param augsAdded The Augemented displays added due to the event
   */
  void open (node metaNode,
	     std::vector<tlp::node> &nodesAdded,
	     std::vector<tlp::edge> &edgesAdded,
	     std::vector<GlABoundingCircle *> &augsAdded);

  /**
   * Function to clean up the display.  It starts at the passed meta-graph
   * and redraws all meta graphs up the branch to the root of the display.
   * \param metaGraph The graph where you start.
   */
  void cleanupDisplay (Graph *metaGraph);

};
#endif
