#include <string>
#include <tulip/ForEach.h>
#include <tulip/DrawingTools.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/DoubleProperty.h>
#include "HierarchyMap.h"
#include "OpenCloseHierarchy.h"
#include "SteeredLayout.h"
using namespace tlp;
using namespace std;

const Coord noPos (0, 0, 0);
const Size noSize (1, 1, 1);

//a function to convert degrees to radians
inline double toRad (double deg) {
  return deg*(M_PI/180.0);
}//end toRad

//a fundtion to convert radians to degrees
inline double toDeg (double rad) {
  return rad*(180.0/M_PI);
}//end toDeg

//a function to rotate by theta in degreesabout a (0, 0) origin
inline Coord rotate (double theta_deg, const Coord pos) {
  Coord result;
  double theta = toRad (theta_deg);
  result.setX (pos.getX()*std::cos(theta) - pos.getY()*std::sin (theta));
  result.setY (pos.getX()*std::sin(theta) + pos.getY()*std::cos (theta));
  result.setZ (0);
  return result;
}//end rotate

//=====================================================================
Coord computeCentre (Graph *subGraph) {
  SizeProperty *sizes = subGraph->getProperty<SizeProperty> ("viewSize");
  LayoutProperty *layout = 
    subGraph->getProperty<LayoutProperty> ("viewLayout");

  //compute the centre of the localGraph by area
  Coord max (-DBL_MAX, -DBL_MAX, 0), min (DBL_MAX, DBL_MAX, 0);
  node curNode;
  forEach (curNode, subGraph->getNodes()) {
    Coord pos = layout->getNodeValue (curNode);
    Size size = sizes->getNodeValue (curNode) / 2.0;
    double rad = sqrt (size.getW()*size.getW() + 
		       size.getH()*size.getH());
    if (max.getX() < pos.getX() + rad)
      max.setX (pos.getX() + rad);
    if (min.getX() > pos.getX() - rad)
      min.setX (pos.getX() - rad);
    if (max.getY() < pos.getY() + rad)
      max.setY (pos.getY() + rad);
    if (min.getY() > pos.getY() - rad)
      min.setY (pos.getY() - rad);
  } //end for each node
  return (max + min)/2.0;
}//end computeCentre

//=====================================================================
void centreByArea (Graph *subGraph, LayoutProperty &subGraphLayout) {
  Coord centre = computeCentre (subGraph);

  //centre the localGraph by area
  node curNode;
  forEach (curNode, subGraph->getNodes())
    subGraphLayout.setNodeValue (curNode,
				 subGraphLayout.getNodeValue (curNode)
				 - centre);

  //update edge bends
  Iterator<edge> *edges = subGraph->getEdges();
  while (edges->hasNext()) {
    edge curEdge = edges->next();
    vector<Coord> bends = subGraphLayout.getEdgeValue (curEdge);
    vector<Coord>::iterator bendsIt = bends.begin();
    for (; bendsIt != bends.end(); ++bendsIt) 
      *bendsIt = *bendsIt - centre;
    subGraphLayout.setEdgeValue (curEdge, bends);
  } delete edges;
}//end centreByArea

//================================================================
//Local function to apply torques to the passed curGraph
void applyTorques (Graph *torqueGraph, DoubleProperty *viewRotation,
		   DoubleProperty *oldRotations) {
  DoubleProperty rotations (torqueGraph);
  string errMsg;
  DataSet torqueData;

  bool success = 
    torqueGraph->computeProperty ("Torques", &rotations, errMsg,
				  NULL, &torqueData);
  if (!success) {
    cerr << "Error calling torques:\n" << endl << errMsg << endl;
    return;
  }//end if

  node curNode;
  forEach (curNode, torqueGraph->getNodes()) {
    oldRotations->setNodeValue (curNode, 
				viewRotation->getNodeValue (curNode));
    viewRotation->setNodeValue (curNode, 
				rotations.getNodeValue (curNode));
  }
}//end applyTorques

//================================================================
void 
OpenCloseHierarchy::updateBoundingCircles (Graph *curHierarchySubGraph,
					   const Coord &cumulativeTrans,
					   double cumulativeRotation,
					   float cumulativeScale) {

  LayoutProperty *hierLayout =
    curHierarchySubGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *hierSize =
    curHierarchySubGraph->getProperty<SizeProperty> ("viewSize");
  DoubleProperty *hierRotation =
    curHierarchySubGraph->getProperty<DoubleProperty> ("viewRotation");
  GraphProperty *hierGraph =
    curHierarchySubGraph->getProperty<GraphProperty> ("viewMetaGraph");
  bool scaling = curHierarchySubGraph->existProperty ("viewScale");

  node curNode;
  forEach (curNode, curHierarchySubGraph->getNodes()) {
    Graph *nodeSubGraph = hierGraph->getNodeValue (curNode);
    if (nodeSubGraph == NULL) continue;
    GlABoundingCircle *metaGraphBoundCircle =
      boundingCircles->getBoundingCircle (curNode);
    if (metaGraphBoundCircle != NULL) {
      Coord curPos = hierLayout->getNodeValue (curNode);
      Size curSize = hierSize->getNodeValue (curNode);
      double curRotation = hierRotation->getNodeValue (curNode);
      if (cumulativeScale > 1e-6) curPos *= cumulativeScale;
      Coord rotCoord = rotate (cumulativeRotation, curPos);
      Coord globalPos = rotCoord + cumulativeTrans;
      curSize *= cumulativeScale;
      metaGraphBoundCircle->update (globalPos, curSize, 
				    metaGraphBoundCircle->getDepth());
      float scaleFactor = 1.0;
      if (scaling) {
	DoubleProperty *viewScale = 
	  curHierarchySubGraph->getProperty<DoubleProperty> ("viewScale");
	scaleFactor = viewScale->getNodeValue (curNode);
      }
      updateBoundingCircles (nodeSubGraph, globalPos,
			     curRotation + cumulativeRotation,
			     scaleFactor*cumulativeScale);
    }//end if 
  }//end forEach
}//end updateBoundingCircle

//================================================================
void 
OpenCloseHierarchy::updateCutLayout (Graph *curHierarchySubGraph,
				     const Coord &cumulativeTrans,
				     double cumulativeRotation,
				     float cumulativeScale) {
  LayoutProperty *hierLayout =
    curHierarchySubGraph->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *hierSize =
    curHierarchySubGraph->getProperty<SizeProperty> ("viewSize");
  DoubleProperty *hierRotation =
    curHierarchySubGraph->getProperty<DoubleProperty> ("viewRotation");
  SizeProperty *openCloseSize =
    openCloseHierarchyGraph->getProperty<SizeProperty> ("viewSize");
  bool scaling = curHierarchySubGraph->existProperty ("viewScale");

  if (curHierarchySubGraph->getAttribute<bool> ("isComplete")) {
    node glyphRoot = curHierarchySubGraph->getAttribute<node> ("root");
    displayLayout->setNodeValue (glyphRoot, cumulativeTrans);
    if (scaling) {
      Size curSize = hierSize->getNodeValue(glyphRoot);
      curSize *= cumulativeScale;
      openCloseSize->setNodeValue (glyphRoot, curSize);
    }
  }
 
  node curNode;
  forEach (curNode, curHierarchySubGraph->getNodes()) {
    Coord curPos = hierLayout->getNodeValue (curNode);
    Size curSize = hierSize->getNodeValue (curNode);
    if (scaling && (cumulativeScale > 1e-6)) 
      curPos *= cumulativeScale;
    double curRotation = hierRotation->getNodeValue (curNode);
    Coord rotCoord = rotate (cumulativeRotation, curPos);
    Coord globalPos = rotCoord + cumulativeTrans;
    Graph *nodeSubGraph = hierarchyMetaGraphs->getNodeValue (curNode);
    if (nodeSubGraph == NULL) {
      displayLayout->setNodeValue (curNode, globalPos);
      if (scaling) {
	Size curSize = hierSize->getNodeValue(curNode);
	curSize *= cumulativeScale;
	openCloseSize->setNodeValue (curNode, curSize);
      }
      continue;
    }//end if

    node groupedNode = hierNodeToDisplayMetaNode.get (curNode.id);
    if (groupedNode.isValid()) {
      displayLayout->setNodeValue (groupedNode, globalPos);
      if (scaling) {
	Size curSize = hierSize->getNodeValue(curNode);
	curSize *= cumulativeScale;
	openCloseSize->setNodeValue (groupedNode, curSize);
      }
      continue;
    }//end if
    else {
      float scaleFactor = 1.0;
      if (scaling) {
	DoubleProperty *viewScale = 
	  curHierarchySubGraph->getProperty<DoubleProperty> ("viewScale");
	scaleFactor = viewScale->getNodeValue (curNode);
      }
      updateCutLayout (nodeSubGraph, globalPos,
		       curRotation + cumulativeRotation, 
		       scaleFactor*cumulativeScale);
    }
  }//end forEach
}//end updateCutLayout

//================================================================
void OpenCloseHierarchy::updateCutLayout (vector<Graph *> &cutDisplayChain) {
  Coord cumulativeTranslation (0, 0, 0);
  Coord cumulativeOldTranslation (0, 0, 0);
  double cumulativeRotation = 0;
  double cumulativeOldRotation = 0;
  Coord nextPosition;
  Coord nextOldPosition;
  double nextRotation;
  double nextOldRotation;
  for (unsigned int i = 0; i < cutDisplayChain.size(); ++i) {
    Graph *fatherGraph = cutDisplayChain[i];
    node fatherHierarchyNode = 
      displaySubIDToHierNode.get (fatherGraph->getId());
    Graph *fatherGraphInHierarchy = 
      hierarchyMetaGraphs->getNodeValue (fatherHierarchyNode);
    SizeProperty *viewSize =
      fatherGraphInHierarchy->getProperty<SizeProperty> ("viewSize");
    LayoutProperty *viewLayout =
      fatherGraphInHierarchy->getProperty<LayoutProperty> ("viewLayout");
    DoubleProperty *viewRotation =
      fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewRotation");
    DoubleProperty *oldRotations = 
      fatherGraphInHierarchy->getProperty<DoubleProperty> ("oldRotation");
    LayoutProperty *oldTranslations =
      fatherGraphInHierarchy->getProperty<LayoutProperty> ("oldTranslation");

    //recentre all the subgraphs one level down
    Iterator<Graph *> *fatherSubGraphs = fatherGraph->getSubGraphs();
    while (fatherSubGraphs->hasNext()) {
      Graph *curSubGraph = fatherSubGraphs->next();
      node curSubGraphHierarchyNode = 
	displaySubIDToHierNode.get (curSubGraph->getId());
      double oldRotation = 
	oldRotations->getNodeValue (curSubGraphHierarchyNode);
      double newRotation = 
	viewRotation->getNodeValue (curSubGraphHierarchyNode);
      Coord newPosition = 
	viewLayout->getNodeValue (curSubGraphHierarchyNode);
      Coord oldPosition =
	oldTranslations->getNodeValue (curSubGraphHierarchyNode);
      Size size = viewSize->getNodeValue (curSubGraphHierarchyNode);

      //only update node positions with composite transform once
      if ((i + 1 < cutDisplayChain.size()) && 
	  cutDisplayChain[i + 1] == curSubGraph) {
	nextPosition = newPosition;
	nextRotation = newRotation;
	nextOldRotation = oldRotation;
	nextOldPosition = oldPosition;
	continue;
      }
      Coord translateBack = oldPosition*(-1) - cumulativeOldTranslation;  
      Coord newGlobalPosition = 
	rotate (cumulativeRotation, newPosition) + cumulativeTranslation;

      double newGlobalRotation = 
	cumulativeRotation - cumulativeOldRotation + 
	newRotation - oldRotation;
      LayoutProperty curSubGraphLayout (curSubGraph);
      node subGraphNode;
      forEach (subGraphNode, curSubGraph->getNodes())
	curSubGraphLayout.setNodeValue (subGraphNode, 
					displayLayout->getNodeValue (subGraphNode));
      centreByArea (curSubGraph, curSubGraphLayout);
      curSubGraphLayout.translate (newGlobalPosition);
      forEach (subGraphNode, curSubGraph->getNodes()) 
	displayLayout->setNodeValue (subGraphNode, 
				     curSubGraphLayout.getNodeValue (subGraphNode));
    } delete fatherSubGraphs;
    
    //do the original nodes of the graph
    node leafNode;
    forEach (leafNode, fatherGraphInHierarchy->getNodes()) {
      if (inputGraph->isElement (leafNode)) {
	Coord relativePos = viewLayout->getNodeValue (leafNode);
	Coord newPosition = rotate (cumulativeRotation, relativePos)
	  + cumulativeTranslation;
	displayLayout->setNodeValue (leafNode, newPosition);
      }//end if
    }//end forEach
    
    //do the edge bends
    edge curEdge;
    forEach (curEdge, fatherGraph->getEdges()) {
      node source = fatherGraph->source (curEdge);
      node target = fatherGraph->target (curEdge);
      node hierarchySource = getHierarchyNode (source);
      node hierarchyTarget = getHierarchyNode (target);
      if (fatherGraphInHierarchy->isElement (hierarchySource) &&
	  fatherGraphInHierarchy->isElement (hierarchyTarget)) {
	edge hierEdge =
	  fatherGraphInHierarchy->existEdge (hierarchySource, hierarchyTarget);
	vector<Coord> edgeBends = viewLayout->getEdgeValue (hierEdge);
	for (unsigned int i = 0; i < edgeBends.size(); ++i)
	  edgeBends[i] = rotate (cumulativeRotation, edgeBends[i]) + 
	    cumulativeTranslation;
	displayLayout->setEdgeValue (curEdge, edgeBends);
      }//end if
    }//end forEach edge

    Coord rotNextPos = rotate (cumulativeRotation, nextPosition);
    cumulativeTranslation += rotNextPos;
    cumulativeRotation += nextRotation;
    cumulativeOldRotation += nextOldRotation;
    cumulativeOldTranslation += nextOldPosition;
  }//end for
}//end updateLayout

//================================================================
void OpenCloseHierarchy::redrawChain (Graph *metaGraph,
				      vector<Graph *> &cutDisplayChain) {
  node metaGraphHierarchyNode = 
    displaySubIDToHierNode.get (metaGraph->getId());
  Graph *metaGraphHierarchyGraph =
    hierarchyMetaGraphs->getNodeValue (metaGraphHierarchyNode);
  Graph *fatherGraph = metaGraph->getSuperGraph();
  node fatherHierarchyNode = displaySubIDToHierNode.get (fatherGraph->getId());
  Graph *fatherGraphInHierarchy = 
    hierarchyMetaGraphs->getNodeValue (fatherHierarchyNode);
  Graph *fatherFatherGraph = fatherGraph->getSuperGraph();
  Graph *fatherFatherInHierarchy = NULL;
  if (fatherFatherGraph != openCloseHierarchyGraph) {
    node fatherFatherNode = 
      displaySubIDToHierNode.get (fatherFatherGraph->getId());
    fatherFatherInHierarchy = 
      hierarchyMetaGraphs->getNodeValue (fatherFatherNode);
  }//end if
    
  
  //compute the bounding box for the current metaGraph and draw one level up
  LayoutProperty *viewLayout =
    fatherGraphInHierarchy->getProperty<LayoutProperty> ("viewLayout");
  SizeProperty *viewSize =
    fatherGraphInHierarchy->getProperty<SizeProperty> ("viewSize");
  SizeProperty *boundingBoxes =
    fatherGraphInHierarchy->getLocalProperty<SizeProperty> ("boundingBox");
  DoubleProperty *viewRotation =
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewRotation");
  DoubleProperty *oldRotations = 
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("oldRotation"); 
  pair<Coord, Coord> metaGraphBoundCircle = 
    tlp::computeBoundingRadius (metaGraphHierarchyGraph, 
				metaGraphHierarchyGraph
				->getProperty<LayoutProperty> ("viewLayout"),
				metaGraphHierarchyGraph
				->getProperty<SizeProperty> ("viewSize"),
				metaGraphHierarchyGraph
				->getProperty<DoubleProperty> ("viewRotation"));
  double radius = (metaGraphBoundCircle.second - 
		   metaGraphBoundCircle.first).norm();
  double diamLength = M_SQRT2*radius;
  Size size (diamLength, diamLength, 1.0);
  viewSize->setNodeValue (metaGraphHierarchyNode, size);
  Size boxSize (2*radius, 2*radius, 1.0);
  boundingBoxes->setNodeValue (metaGraphHierarchyNode, boxSize);

  drawSteeredNode (fatherGraphInHierarchy, fatherFatherInHierarchy,
		   fatherGraph);
  applyTorques (fatherGraphInHierarchy, viewRotation, oldRotations);
  centreByArea (fatherGraphInHierarchy, *viewLayout);

  if (fatherGraph->getSuperGraph() != openCloseHierarchyGraph)
    redrawChain (fatherGraph, cutDisplayChain);
  cutDisplayChain.push_back (fatherGraph);
}//cleanupDisplay

//================================================================
float scaleFactor (double layoutRad, double nodeRad) {
  return (float)nodeRad/(float)layoutRad;
}
//================================================================
void OpenCloseHierarchy::scaleDown (Graph *metaGraph) {
  node metaGraphHierarchyNode = 
    displaySubIDToHierNode.get (metaGraph->getId());
  Graph *metaGraphHierarchyGraph = 
    hierarchyMetaGraphs->getNodeValue (metaGraphHierarchyNode);
  Graph *fatherGraph = metaGraph->getSuperGraph();
  node fatherGraphHierarchyNode = 
    displaySubIDToHierNode.get (fatherGraph->getId());
  Graph *fatherGraphInHierarchy = 
    hierarchyMetaGraphs->getNodeValue (fatherGraphHierarchyNode);
  DoubleProperty *metaGraphHierarchyGraphScale =
    metaGraphHierarchyGraph->getProperty<DoubleProperty> ("viewScale");
  metaGraphHierarchyGraphScale->setAllNodeValue (1.0);

  LayoutProperty *viewLayout =
     fatherGraphInHierarchy->getProperty<LayoutProperty> ("viewLayout");
  DoubleProperty *viewScale =
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewScale");
  SizeProperty *viewSize =
    fatherGraphInHierarchy->getProperty<SizeProperty> ("viewSize");
  DoubleProperty *viewRotation =
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("viewRotation");
  DoubleProperty *oldRotations = 
    fatherGraphInHierarchy->getProperty<DoubleProperty> ("oldRotation"); 
  SizeProperty *boundingBoxes =
    fatherGraphInHierarchy->getLocalProperty<SizeProperty> ("boundingBox");
  
  pair<Coord, Coord> metaGraphBoundCircle = 
    tlp::computeBoundingRadius (metaGraphHierarchyGraph, 
				metaGraphHierarchyGraph
				->getProperty<LayoutProperty> ("viewLayout"),
				metaGraphHierarchyGraph
				->getProperty<SizeProperty> ("viewSize"),
				metaGraphHierarchyGraph
				->getProperty<DoubleProperty> ("viewRotation"));
  double radius = (metaGraphBoundCircle.second - 
		   metaGraphBoundCircle.first).norm();
  Size nodeSizeEst = viewSize->getNodeValue (metaGraphHierarchyNode) / 2.0;
  double nodeRad = sqrt (nodeSizeEst.getW()*nodeSizeEst.getW() + 
			 nodeSizeEst.getH()*nodeSizeEst.getH());
  Size boxSize (2*nodeRad, 2*nodeRad, 1.0);
  boundingBoxes->setNodeValue (metaGraphHierarchyNode, boxSize);
  viewScale->setNodeValue (metaGraphHierarchyNode, 
			   scaleFactor (radius, nodeRad));
  applyTorques (fatherGraphInHierarchy, viewRotation, oldRotations);
  centreByArea (fatherGraphInHierarchy, *viewLayout);
  updateCutLayout (hierarchy, noPos, 0, 1.0);
  updateBoundingCircles (hierarchy, noPos, 0, 1.0);
}

//================================================================
void OpenCloseHierarchy::cleanupDisplay (Graph *metaGraph) {
  vector<Graph *> cutDisplayChain;
  redrawChain (metaGraph, cutDisplayChain);
  updateCutLayout (hierarchy, noPos, 0, 1.0);
  updateBoundingCircles (hierarchy, noPos, 0, 1.0);
}
