#include <string>
#include <tulip/TreeTest.h>
#include <tulip/ForEach.h>
#include <tulip/DrawingTools.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/DoubleProperty.h>
#include "HierarchyMap.h"
#include "SteeredLayout.h"
using namespace tlp;
using namespace std;

//================================================================
//Private helper function to create a new edge order table
void setEdgeOrderTable (Graph *curSubGraph) {
  EdgeOrderTable *newEdgeOrderTable = new EdgeOrderTable();
  node curNode;
  forEach (curNode, curSubGraph->getNodes()) {
    vector<edge> curEdgeList;
    edge curEdge;
    forEach (curEdge, curSubGraph->getInOutEdges(curNode))
      curEdgeList.push_back (curEdge);
    newEdgeOrderTable->set (curNode.id, curEdgeList);
  }//end forEach
  curSubGraph->setAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE,
					       newEdgeOrderTable);
}//end setEdgeOrderTable

//================================================================
//private helper function to set the order of the edges around each
//node consistantly.
void setEdgeOrder (Graph *curSubGraph) {
  EdgeOrderTable *curEdgeOrderTable =
    curSubGraph->getAttribute<EdgeOrderTable *>(EDGE_ORDER_ATTRIBUTE);
  node curNode;
  forEach (curNode, curSubGraph->getNodes()) {
    const vector<edge> &edgeOrder = curEdgeOrderTable->get (curNode.id);
    curSubGraph->setEdgeOrder (curNode, edgeOrder);
  }//end forEach
}//end setEdgeOrder

//================================================================
inline double computeRad (const Size &curSize) {
  return sqrt (curSize.getW()*curSize.getW()/4.0 + 
	       curSize.getH()*curSize.getH()/4.0);
}//end computeRad

//================================================================
//set up the GEM dataset edge length attributes
void gemEdgeLength(Graph *curSubGraph, DoubleProperty &edgeWeight) {
  SizeProperty *viewSize =
    curSubGraph->getProperty<SizeProperty> ("viewSize");
  edge curEdge;
  forEach (curEdge, curSubGraph->getEdges()) {
    node source = curSubGraph->source (curEdge);
    node target = curSubGraph->target (curEdge);
    Size sourceSize = viewSize->getNodeValue (source);
    Size targetSize = viewSize->getNodeValue (target);
    double weight = computeRad (sourceSize) + computeRad (targetSize);
    edgeWeight.setEdgeValue (curEdge, weight);
  }//end forEach
}

//================================================================
//set up the initial layout
void initializeLayout (Graph *curSubGraph, LayoutProperty &initLayout) {
  LayoutProperty *viewLayout =
    curSubGraph->getProperty<LayoutProperty> ("viewLayout");
  node curNode;
  forEach (curNode, curSubGraph->getNodes())
    initLayout.setNodeValue (curNode, 
			     viewLayout->getNodeValue (curNode));
}//end initializeLayout

//================================================================
void rootBiconnected (Graph *curSubGraph) {
  node curNode, maxDegNode;
  unsigned int maxDeg = 0;
  forEach (curNode, curSubGraph->getNodes()) {
    if (curSubGraph->deg (curNode) > maxDeg) {
      maxDeg = curSubGraph->deg (curNode);
      maxDegNode = curNode;
    }//end if
  }//end forEach
  TreeTest::makeRootedTree (curSubGraph, maxDegNode);
}

//================================================================
void rootTree (Graph *curSubGraph, Graph *fatherSubGraph) {
  node curNode, rootNode;
  Graph *root = curSubGraph->getRoot();
  bool done = false;
  forEach (curNode, curSubGraph->getNodes()) {
    node oppositeNode;
    forEach (oppositeNode, root->getInOutNodes (curNode)) {
      if (fatherSubGraph->isElement (oppositeNode)) {
	rootNode = curNode;
	done = true;
	breakForEach;
      }//end if
    }//end forEach
    if (done) breakForEach;
  }//endForEach
  TreeTest::makeRootedTree (curSubGraph, rootNode);
}

//================================================================
bool drawSteeredNode (Graph *curSubGraph, Graph *fatherSubGraph,
		      Graph *displaySubGraph) {
  LayoutProperty *curSubGraphLayout =
    curSubGraph->getProperty<LayoutProperty> ("viewLayout");
  DataSet topoLayoutDrawData;
  string defaultDrawOperator = string (TREE_IDENT) +
    ";Bubble Tree;Hierarchical Tree (R-T Extended);" + 
    string (CONNECTED_IDENT) + 
    ";Connected Componnent Packing;Connected Componnent Packing;" +
    string (BICONNECTED_IDENT) +
    ";Bubble Tree;Hierarchical Tree (R-T Extended);" + 
    string (EMBEDDER_IDENT) +
    ";High Dimensional Embedder;High Dimensional Embedder;" +
    string (STRENGTH_IDENT) + ";GEM (Frick);GEM (Frick);" +
    string (UNKNOWN_IDENT) + ";GEM (Frick);GEM (Frick);" +
    string (NEAR_COMPLETE_IDENT) + ";Circular;Circular;" +
    string (LGL_IDENT) + ";LGL Area;LGL Area;";
  
  /*string defaultDrawOperator = string (TREE_IDENT) +
    ";GEM (Frick);GEM (Frick);" + 
    string (CONNECTED_IDENT) + 
    ";Connected Componnent Packing;Connected Componnent Packing;" +
    string (BICONNECTED_IDENT) +
    ";GEM (Frick);GEM (Frick);" + 
    string (EMBEDDER_IDENT) +
    ";GEM (Frick);GEM (Frick);" +
    string (STRENGTH_IDENT) + ";GEM (Frick);GEM (Frick);" +
    string (UNKNOWN_IDENT) + ";GEM (Frick);GEM (Frick);" +
    string (NEAR_COMPLETE_IDENT) + ";GEM (Frick);GEM (Frick);" +
    string (LGL_IDENT) + ";GEM (Frick);GEM (Frick);";
  */
  topoLayoutDrawData.set (DATASET_DRAW_OPERATOR, defaultDrawOperator);
  topoLayoutDrawData.set<bool> (DATASET_CLUSTER_BUSTER, false);
  topoLayoutDrawData.set<bool> (DATASET_FAST_OVERLAPS, true);
  topoLayoutDrawData.set<bool> (DATASET_TORQUES, false);
  topoLayoutDrawData.set<bool> (DATASET_DRAW_THIS, true);

  //save the edge orders if they don't have them already.  And set
  //the edge orders around each node for consistant drawing
  EdgeOrderTable *curEdgeOrderTable =
    curSubGraph->getAttribute<EdgeOrderTable *> (EDGE_ORDER_ATTRIBUTE);
  if (curEdgeOrderTable == NULL) 
    setEdgeOrderTable (curSubGraph);
  setEdgeOrder (curSubGraph);
  
  const DataSet &attributes = curSubGraph->getAttributes(); 
  if (isType (curSubGraph, EMBEDDER_IDENT) && 
      !attributes.exist (HDE_ROOT_ATTRIBUTE))
    curSubGraph->setAttribute<node> (HDE_ROOT_ATTRIBUTE, 
				     curSubGraph->getOneNode());
  
  else if (isType (curSubGraph, BICONNECTED_IDENT))
    rootBiconnected (curSubGraph);
  
  else if (isType (curSubGraph, TREE_IDENT)) {
    if (fatherSubGraph != NULL)
      rootTree (curSubGraph, fatherSubGraph);
    else
      rootBiconnected (curSubGraph);
  }//end if

 
  //if we have a graph where we are going to use GEM
  DoubleProperty edgeLengths (curSubGraph);
  LayoutProperty initLayout (curSubGraph);
  edgeLengths.setAllEdgeValue (1.0);
  if (isType (curSubGraph, UNKNOWN_IDENT) ||
      isType (curSubGraph, STRENGTH_IDENT)) {
    gemEdgeLength(curSubGraph, edgeLengths);
    //topoLayoutDrawData.set ("edge length", &edgeLengths);
    bool hasLayout = curSubGraph->getAttribute<bool> (HAS_LAYOUT_ATTRIBUTE);
    if (hasLayout) {
      initializeLayout (curSubGraph, initLayout);
      topoLayoutDrawData.set ("initialisation", &initLayout);
    }//end if
  }//end if

  string errMsg;
  LayoutProperty hierLayout (curSubGraph);
  bool success = 
    curSubGraph->computeProperty ("TopoLayout", &hierLayout, errMsg,
				  NULL, &topoLayoutDrawData);
  if (!success) {
    cerr << "Error in calling TopoLayout:" << endl << errMsg << endl;
    return false;
  }//end if
  
  node curNode;
  forEach (curNode, curSubGraph->getNodes()) 
    curSubGraphLayout->setNodeValue (curNode, 
				     hierLayout.getNodeValue (curNode));
  edge curEdge;
  forEach (curEdge, curSubGraph->getEdges())
    curSubGraphLayout->setEdgeValue (curEdge,
				     hierLayout.getEdgeValue (curEdge));
  curSubGraph->setAttribute<bool> (HAS_LAYOUT_ATTRIBUTE, true);
  return success;
}
